##########################################################################
#
# FILE:
#       cmdline.pm
# 
# DESCRIPTION: 
#       Functions to handle commandline parsing
#
# AUTHORS:
#       Paul Russo
#	Marc Mengel
#	Lars Rasmussen
#       Margaret Votava
#
#       Fermilab Computing Division
#       Batavia, Il 60510, U.S.A.                                              
#
# MODIFICATIONS:
#	28-jul-1998, llc, add $ERRWIDTH and cmdline_setwidth;
#	                  modify cmdline_log so that it handles
#	                  lines of arbitrary initial width and
#                         breaks them into chunks no larger than
#                         $ERRWIDTH
#       dd-mmm-yyyy, NN, first
#
##########################################################################

package cmdline;
use updgbl;

require Exporter;

@ISA = qw(Exporter);

##########################################################################
# Exported Functions
#
##########################################################################

@EXPORT = qw ( cmdline_process_options );

##########################################################################
# Global and local variables
#
##########################################################################

#-------------------------------------------------------------------------
# $status = cmdline_process_options( $control, \@args, \%flags, \@argv );
#
# Process the command line arguments and options, returning the arguments
# and options found.
#
# Input : $control -- Control string.  Use getopts(3) format.
#
# Output: $args  -- Reference to an array containing the arguments
#                   from the command line.
#         $flags -- Reference to a hash containing summary
#                   of command line options.
#
# Return: $status -- Always true.
#

sub cmdline_process_options {

    #
    #  Passed arguments.
    #

    my $control = shift() . " ";    #  Option parsing control string.

                    #  The extra space is for the
                    #  convenience of the argument checker.

    my $outargs = shift();      #  Array ref for command line arguments.
    my $outflags = shift();     #  Hash ref for command line options.
        my $inargs = shift();

    #
    #  Private variables.
    #

    my $fc;             #  First character (option letter).
    my $rest;           #  Rest of option after first character.
    my $x;              #  Index into control string.

    #
    #  Check passed arguments.
    #

    if (       not defined( $outargs )
        or not ref( $outargs )
        or ( ref( $outargs ) ne "ARRAY" )
    ) {
        die( "Invalid arguments!, stopped" );
    }

    if (       not defined( $outflags )
        or not ref( $outflags )
        or ( ref( $outflags ) ne "HASH" )
    ) {
        die( "Invalid arguments!, stopped" );
    }

    #
    #  Loop over all the arguments.
    #

    while ( @{$inargs} ) {

        #
        #  Get the next argument.
        #

        $_ = shift( @{$inargs} );

        #
        #  Is it an option?
        #

        if ( not /^-(.)(.*)/ ) {

            #
            #  Not an option, we have an argument.  Save away,
            #  so we can return it to our caller.
            #

            push( @{$outargs}, $_ );

            #
            #  Move on to next argument.
            #

            next;

        }

        #
        #  Ok, we had an option.  Pull it out of the argument and
        #  keep the rest for option clustering and possibly for use
        #  as an argument to the option.
        #

        ( $fc, $rest ) = ( $1, $2 );

        #
        #  Look up the option letter in the passed control string.
        #

        $x = index( $control, $fc );

        #
        #  Did we match an option in the control string?
        #

        if ( -1 == $x ) {

            #
            #  Nope, crash and burn.
            #

            die( "Unknown option: -$fc, stopped" );

        } else {

            #
            #  Ok, we have a legal option.  Does it
            #  require an argument?
            #

            #
            #  N.B.:  This is where we needed the extra
            #         space added to the control string
            #         ( so we can go one char past nominal end).
            #

            if ( substr( $control, $x + 1, 1 ) eq ":" ) {

                #
                #  Ok, we need an argument for the option.
                #  Use what was included with the option,
                #  if anything, otherwise use the next
                #  argument.
                #

                if (        defined( $rest )
                    and ( $rest ne "" )
                ) {
                    $_ = $rest;
                } else {
                    $_ = shift( @{$inargs} );
                }

                if ( not defined( $_ ) ) {
            die( "Option -$fc requires an argument!, stopped" );
                }

                #
                #  We got the option's required argument.
                #  Put it into the flags hash.
                #

                $outflags->{"opt_$fc"} = $_;

            } else {

                #
                #  We got a legal option which needs no
                #  argument.  Update the count for the option
                #  in the flags hash.
                #

                #
                #  Handle the -? option in a special way,
                #  since '?' is invalid in variable names.
                #
                
                if ( $fc eq "?" ) {
                    $fc = "help";
                }

                if ( not exists( $outflags->{"opt_$fc"} ) ) {
                    $outflags->{"opt_$fc"} = 1;
                } else {
                    $outflags->{"opt_$fc"}++;
                }

                #
                #  Allow option clustering.
                #

                if (        defined( $rest )
                    and ( $rest ne "" )
                ) {
                    unshift( @{$inargs}, "-" . $rest );
                }

            }

        }

    }

    #
    #  Return to caller with success status.
    #

    return( 1 );

}

