##########################################################################
#
# FILE:
#       updchk.pm
# 
# DESCRIPTION: 
#       This module contain routines that are used to check consistency
#       prerequisites.
#       (for more info, see pod at the end).
#
# AUTHORS:
#       Marc Mengel
#       Lars Rasmussen
#       Margaret Votava
#       Don Walsh
#
#       Fermilab Computing Division
#       Batavia, Il 60510, U.S.A.                                              
#
# MODIFICATIONS:
#       dd-mmm-yyyy, NN, first
#
##########################################################################
package updchk;

use upderr;
use strict;
use integer;
use Net::Ping;
use Exporter();
use vars qw( @ISA @EXPORT );
@ISA = qw( Exporter );

# export this !
@EXPORT = qw (
 	updchk_instance_spec
	updchk_manifest
	updchk_prerequisite
 	updchk_size
	);

##########################################################################
# BEGIN
#
# Check if neccesary environment variables are present, and that the 
# distribution node is valid.
#
# Any check there only have to be done once at load time can be added
# here.
# 
sub BEGIN {

  # checking for a valid local ups environment.
  # should we check for a valid ups version ?

  die "UPD_DIR is not defined ... we are quitting.\n" 
      unless ( defined $ENV{UPD_DIR} );
  die "UPS_DIR is not defined ... we are quitting.\n" 
      unless ( defined $ENV{UPS_DIR} );
#  die "\$UPS_DIR/bin/ups does not exist ... we are quitting.\n" 
#      unless ( -e "$ENV{UPS_DIR}/bin/ups" );
  die "PRODUCTS is not defined ... we are quitting.\n" 
      unless ( defined  $ENV{PRODUCTS} );
}

##########################################################################
# Exported functions
#

#-------------------------------------------------------------------------
# status = updchk_instance_spec( \%instance_spec, $what )
#
# Checks the instance spec to make sure all necessary fields are filled for
# a specific routine.
#
# Input : hash, instance_spec
#         scalar, what, how to check instance spec (could be a routine name)
# Output: none
# Return: scalar, status
#
sub updchk_instance_spec {
(my $inst_spec, $_) = @_;
upderr_log(4,"args ($inst_spec,$_)\n");
#
# for all routines
#
unless ( defined $inst_spec->{product} )
   {
   upderr_log(0,"ERROR: Product not specified.\n");
   return 0;
   }
unless ( defined $inst_spec->{flavor} )
   {
   upderr_log(0,"ERROR: Flavor not specified for $inst_spec->{product}.\n");
   return 0;
   }
unless ( defined $inst_spec->{qualifiers} )
   {
   upderr_log(0,"ERROR: qualifiers not specified for $inst_spec->{product} \n");
   return 0;
   }
#
# check for a unchain call
#
if ( /^unchain/ ) {
    return 1; # all checked in above test
    unless ( defined $inst_spec->{chain} )
       {
       upderr_log(0,"ERROR: chain not specified for $inst_spec->{product} \n");
       return 0;
       }
   }
unless ( defined $inst_spec->{version} )
   {
   upderr_log(0,"ERROR: version not specified for $inst_spec->{product} \n");
   return 0;
   }
#
# check for a unique instance
#
if ( /^unique/ ) {
    return 1; # all checked in above test
   }
if ( /^updchk_depend/ )
   {
   unless ( defined $inst_spec->{foundby} )
      {
      upderr_log (0,"ERROR: foundby path not specified for " .
         "$inst_spec->{product} \n");
      return 0;
      }
   unless ( defined $inst_spec->{database} )
      {
      upderr_log (0,"ERROR: database not specified for " .
         "$inst_spec->{product} \n");
      return 0;
      }
   return 1;
   }
if ( /^updusr_get_declare_proddir/ )
   {
   unless ( defined $inst_spec->{unwind_dir} )
      {
      upderr_log (0,"ERROR: unwind_dir not specified for " .
         "$inst_spec->{product} \n");
      return 0;
      }
   return 1;
   }

if ( /^updups_declare/ ) {
   # mengel -- table file only is legal...
   # unless (  defined $inst_spec->{prod_dir} ) { 
   # upderr_log (0,"ERROR: updups_declare 'prod_dir' not specified for " .
   #      "$inst_spec->{product} \n");
   # return 0;    
   #}
  return 1;
}

upderr_log (0,"ERROR: invalid subroutine name \"$_\" .\n");
return 0;
}

#-------------------------------------------------------------------------
# status = updchk_manifest( $proddir )
#
# Looks for a $prod_dir/.manifest file, and if it finds one, checks it 
# against the file actually there.
#
# Input : scalar, $proddir
# Output: none
# Return: scalar, status
#
sub updchk_manifest {
my ($proddir,$prodname) = @_;

$prodname = "" if (!defined $prodname);
my $MANFD;			# file descriptor for manifest file
my $fname="$proddir/.manifest"; # actual file name
my $errflg = 1;			# error flag
my $count;			# missing file count

upderr_log(3,"updchk_manifest($proddir)\n");
return 1 unless ( -f "$fname.$prodname" || -f $fname); 	# doesn't exist, that's ok 
unless (open(MANFD,"$fname.$prodname")||open(MANFD,$fname))
   {
   upderr_log(0,"ERROR: Could not open manifest file $fname.\n");
   return 0;
   }
$count = 0;
while (<MANFD>)
   {
   next unless ( /(\d)+\s+(\S+)/ );	# get checksum, filename
   $fname = "$proddir/$2";
   unless ( -e $fname )
      {
      $count++;
      $errflg = 0;
      if ($count < 5) {
          upderr_log(0,"ERROR: \"$fname\" does not exist in local product " .
                    "instance.\n");
         }
      }
   }
if ($count >= 5) {
   upderr_log(0, "...\nA total of $count files were missing\n");
   }
close (MANFD);
return $errflg;
}


#-------------------------------------------------------------------------
# status = updchk_prerequisite( $distnode )
#
# Checks the distribution node to see if it's valid
#
# Input : scalar, distnode, distribution node name
# Output: none
# Return: scalar, status
#
sub updchk_prerequisite {

my ($distnode) = @_;

upderr_log(3,"updchk_prerequisite($distnode)\n");
unless ( pingecho ($distnode,10))	# check if node is up
   {
   upderr_log (0,"ERROR: node \"$distnode\" is unreachable \n");
   return 0;
   }
return 1;				# success
}

#-------------------------------------------------------------------------
# status = updchk_size ( )
#
#
# Input : 
# Output: none
# Return: scalar, status
#
sub updchk_size {

upderr_log(3,"updchk_size()\n");
print "updchk_size: not done yet\n";
return 0;
}

##########################################################################
# Private functions
#

1;

__END__

=head1 NAME

updchk - This module contain routines that are used to check consistency 
prerequisites

=head1 SYNOPSIS

<use it this way>

=head1 DESCRIPTION
This module contain routines that are used to check consistency 
prerequisites

<description>

=head1 BUGS

<bugs>

=cut
