##########################################################################
#
# FILE:
#       updchk2.pm
# 
# DESCRIPTION: 
#       This module contain routines that are used to check dependencies
#	Ideally this would have been in updchk.pm, but we had a problem
#	with circular dependencies updchk -> updups and updups_updchk.
#       (for more info, see pod at the end).
#
# AUTHORS:
#       Marc Mengel
#       Lars Rasmussen
#       Margaret Votava
#       Don Walsh
#
#       Fermilab Computing Division
#       Batavia, Il 60510, U.S.A.                                              
#
# MODIFICATIONS:
#       dd-mmm-yyyy, NN, first
#
##########################################################################
package updchk2;

use upderr;
use updups;
use updchk;
use upduti;
use updgbl;
use upduti;
use strict;
use integer;
use Exporter();
use vars qw( @ISA @EXPORT );
@ISA = qw( Exporter );

# export this !
@EXPORT = qw (
	updchk_depend
	);

##########################################################################
# Exported functions
#

#-------------------------------------------------------------------------
# status = updchk_depend ( \@local_depend, \@dist_depend, *FD )
#
# checks local dependency list against distribution dependency list.
# If lists are different, outputs the appropriate ups commands on the
# local node to make them the same.
#
# Input : reference, $loc_depend
#       : reference, $dist_depend
#       : filehandle, $FD
# Output: none
# Return: scalar, status
#
sub updchk_depend {
my ($loc_list ,         # how depend look on local node
    $dist_list,         # how depend look on dist node
    $FD                 # where to output differences
) = @_;

my $loc_inst={};        # instance in local node list
my $dist_inst='';       # instance in distribution node list
my $errflg=1;           # our error flag
my ($dcount,$lcount);   # number of items in list
my $nloop;              # number of times through loop
my %tmp_inst;           # local instance for ups_exist
my $exist;              # T/F exist flag
my $printonce = 0;	# T/F flag

upderr_log(1,"updchk_depend($loc_list,$dist_list,$FD)\n");

#
# make sure lists sizes are the same. Not necessarily an error.
#
$dcount = scalar(@$dist_list);
$lcount = scalar(@$loc_list);

# Don't complain about this; its almost always true if there are
# products unchained, and we tell them what chains to make anyway...
#upderr_log(0,"WARNING: lists aren't the same size dist node = $dcount, " .
#             "local node = $lcount\n") unless ($dcount == $lcount);
#
# loop through all dependencies on the distribution node
#
DIST_LOOP: foreach $dist_inst (@$dist_list)
   {
   return 1 unless (updchk_instance_spec(\%$dist_inst,"updchk_depend"));
   $nloop = 0;
#
# loop through the local node dependencies looking for a match
#
   LOCAL_LOOP: foreach $loc_inst (@$loc_list)
      {
      $nloop++;
      return 1 unless (updchk_instance_spec(\%$loc_inst,"updchk_depend"));
      if ( $dist_inst->{product} eq $loc_inst->{product} )
         {
#
# our flavors needs to match, else an error
#
         unless ($dist_inst->{flavor} eq $loc_inst->{flavor} )
            {
            upderr_log (0, "Warning: For product \"$loc_inst->{product}\"".
               "local node flavor $loc_inst->{flavor} does not match ".
               "distribution node flavor $dist_inst->{flavor} \n");
            next DIST_LOOP;
            }
#
# our qualifiers need to match, else an error
#
         unless ( $dist_inst->{qualifiers} eq $loc_inst->{qualifiers} )
            {
            upderr_log (0, "Warning: For product \"$loc_inst->{product}\"".
               "local node qualifiers $loc_inst->{qualifiers} do not match ".
               "distribution node qualifiers $dist_inst->{qualifiers} \n");
            next DIST_LOOP;
            }
#
# We got this dependency though the same route and the versions are the
# same.
#
         next DIST_LOOP if (($dist_inst->{foundby} eq $loc_inst->{foundby}) and
                            ($dist_inst->{version} eq $loc_inst->{version}));
#
# We got the distribution node dependency by an explicit version. The local
# node had better have the same version, or an error
#
         if ($dist_inst->{foundby} eq "")
            {
            unless ( $dist_inst->{version} eq $loc_inst->{version} )
               {
               upderr_log (0, "Warning: For product \"$loc_inst->{product}\"".
                  "local node version $loc_inst->{version} does not match ".
                  "distribution node version $dist_inst->{version} \n");
               }
            next DIST_LOOP;
            }
	 last LOCAL_LOOP;
         }    # product name matched
      }    # LOCAL_LOOP

# if we got this far, distribution dependency must be a 
# chain or there is an error
# -----------------------------------------------------
if ($dist_inst->{foundby} eq "")
   {
   upderr_log(0,"WARNING: product $dist_inst->{product} not in local " .
      "dependency list\n");
   next DIST_LOOP;
   }

# see if this local node has this instance
# Note that we have to get the information about the specific instance from 
# the distribution node data, but that the database comes from the local 
# node ...
# --------------------------------------------------------------------------
   %tmp_inst = %$dist_inst;

   if ($nloop == $lcount)			# product not in local list
      { delete $tmp_inst{database}; }
   else						# product in local list
      { $tmp_inst{database} = $loc_inst->{database}; }
   
   unless ( updups_exist(0,\%tmp_inst,\$exist) )
      {
      upderr_log (0, "Error: updups_exist error for product " .
             "\"$tmp_inst{product}\" \n");
      $errflg = 0;
      next DIST_LOOP;
      }
   if ( $exist )
      {
	  # make sure we have a database
	  # the _hack1 is here to work around a ups v4_0 bug
	  updups_find(0,\%tmp_inst,'_hack1', 'database');
      }
   else
      {
      upderr_log (0, "Error: need to declare chain for product, ".
                "but it doesn't exist. \n");
      upduti_print_instance_spec(0,\%tmp_inst);
      $errflg = 0;
      next DIST_LOOP;
      }

# Let's output the ups command to make the local instance have the correct 
# chain. 
#
   unless ($printonce)
      {
      upderr_log(0,"Execute the following to resolve chains: \n");
      $printonce = 1;
      }
   upderr_get(); upderr_clear();		# flush error stack
   print $FD "ups declare -f $dist_inst->{flavor} ".
             "-q \"$dist_inst->{qualifiers}\" -g $dist_inst->{foundby} ".
             "$dist_inst->{product} $dist_inst->{version} ";
   if ($tmp_inst{database} eq "" )
       { print $FD "\n"; }
   else
       { print $FD "-z $tmp_inst{database}\n"; }
   }    # DIST_LOOP
return $errflg;
}


##########################################################################
# Private functions
#

# If standalone, run tests
if ( $0 =~ /updchk2.pm$/ ) 
{
#unless ( $#ARGV == -1 ) upderr_setverbose(10);
my %a_inst = (
           product     => "exmh",
           flavor      => "IRIX",
           version     => 'v1_3',
           qualifiers  => '',
           database    => "blah",
           foundby     => "current"
           );
my %b_inst = (
           product     => "tcl",
           flavor      => "IRIX",
           version     => 'v2_2',
           qualifiers  => '',
           foundby     => "current",
           database    => "blah",
           );
my %c_inst = (
           product     => "tk",
           flavor      => "IRIX",
           version     => 'v3_0',
           qualifiers  => '',
           foundby     => "current",
           database    => "blah",
           );
my %d_inst = (
           product     => "exmh",
           flavor      => "IRIX",
           version     => 'v1_2',
           qualifiers  => '',
           foundby     => "current",
           database    => "blah",
           );
my %e_inst = (
           product     => "tcl",
           flavor      => "IRIX",
           version     => 'v2_2',
           qualifiers  => '',
           foundby     => "current",
           database    => "blah",
           );
my %f_inst = (
           product     => "tk",
           flavor      => "IRIX",
           version     => 'v3_0',
           qualifiers  => '',
           foundby     => "current",
           database    => "blah",
           );
my %g_inst = (
           product     => "ghostview",
           flavor      => "IRIX",
           version     => 'v3_0',
           qualifiers  => '',
           foundby     => "current",
           database    => "blah",
           );


my @loc_list =  (\%a_inst, \%b_inst, \%c_inst);
my @dist_list = (\%d_inst, \%e_inst, \%f_inst, \%g_inst);

my $ret_val;
my $proddir = ".";
my $err;

$g_flags{opt_h} = "blah";
$ret_val = updchk_prerequisite( );
upderr_get(\$err);
print "main: output from updchk_prerequisite = $ret_val\n $err";

$g_flags{opt_h} = "upd";
$ret_val = updchk_prerequisite( );
upderr_get(\$err);
print "main: output from updchk_prerequisite = $ret_val\n $err";

#$ret_val = updchk_depend( \@loc_list, \@dist_list, \*STDOUT);
upderr_get(\$err);
print "main: output from updchk_depend: = $ret_val\n $err";

$ret_val = updchk_manifest($proddir,"ghostview"); # XXX is ghostview right? mengel 
upderr_get(\$err);
print "main: output from updchk_manifest: = $ret_val\n $err";
}

1;

__END__

=head1 NAME

updchk - This module contain routines that are used to check consistency 
prerequisites

=head1 SYNOPSIS

<use it this way>

=head1 DESCRIPTION
This module contain routines that are used to check consistency 
prerequisites

<description>

=head1 BUGS

<bugs>

=cut
