##########################################################################
#
# FILE:
#       updcmd.pm
# 
# DESCRIPTION: 
#	These are the actual upd commands: install, update, and fetch
#       (for more info, see pod at the end).
#
# AUTHORS:
#       Marc Mengel
#       Lars Rasmussen
#	Paul Russo
#       Margaret Votava
#       Don Walsh
#
#       Fermilab Computing Division
#       Batavia, Il 60510, U.S.A.                                              
#
# MODIFICATIONS:
#       dd-mmm-yyyy, NN, first
#
##########################################################################
package updcmd;

use updgbl;
use updchk2;
use updchk;
use updups;
use updupr;
use updstb;
use updui;
use upderr;
use updxfr;
use upduti;
use File::Path;
use File::Copy;

use strict;
use integer;
use Exporter();
use vars qw( @ISA @EXPORT);
@ISA = qw(Exporter);

# export this !
@EXPORT = qw (
	      updcmd_install
              updcmd_update
              updcmd_fetch
	      updcmd_add_mod_product
	      updcmd_cloneproduct
	      updcmd_delproduct
	      updcmd_chainproduct
	      updcmd_move_archive_file
	      updcmd_move_table_file
	      updcmd_move_ups_dir
	      updcmd_get_proprietarylist
	      );

##########################################################################
# Exported functions
#

#-------------------------------------------------------------------------
# status = updcmd_install ( \%install_spec )
#
# This is the routine to install a new product instance and/or dependencies.
#
# Input : reference, %install_spec	instance spec to install
#       : scalar, dist_db		distribution database
# Output: none
# Return: scalar, status
#

sub updcmd_install
{
    my ($install_spec) = @_;
    my $errstr;			# error string
    my @dist_list;			# distribution node dependencies
    my @loc_list;			# local node dependencies
    my $inst_spec;			# instance spec
    my $exist;			# existance flag
    my $ninst = 0;			# number of instances to install
    my @inst_list;			# list of instances to install
    my %has_overlays;		# defined for prodname if it has overlays
    my %name_to_spec;		# hash to find named products fast
    my %parent_prod;		# hash of parent products
    my $dist_url;			# distribution url (a hash reference)
    my $dist_db;
    my $status;			# subroutine return status
    my $savedb;			# temporary slot for database
    my $local_spec;		# local instance spec of product


    #
    &upderr_log(1,"updcmd_install($install_spec)\n");

    if ( (defined($g_flags{opt_r}) || defined($g_flags{opt_M}) || 
		defined($g_flags{opt_m}) || defined($g_flags{opt_U})) &&
		!defined($g_flags{opt_j})) {

        upderr_log(0,"Notice: -j flag turned on due to location overrides");
        $g_flags{opt_j} = 1;
    }

    return 0 unless 
       updcmd_split_init(\$dist_url,$install_spec,\@dist_list,\$dist_db);

    upderr_log(3,"after init:\n\tdist_node is $dist_url->{node},\n\tdist_list is ".
		  join(":", @dist_list). "\n\tdist_db is $dist_db\n" );


    # for each product in the dependency list
    # ---------------------------------------
    upderr_log(3, "going through depend list $#dist_list $dist_url->{node} \n");
    foreach $inst_spec (@dist_list)  {

        upderr_log (3,"looking at $inst_spec->{product}.\n");
        updcmd_split_bookkeeping( $dist_url, $inst_spec, 
			\%has_overlays, \%name_to_spec, \%parent_prod, \$exist, 
			\$local_spec);

        if ( $exist ) {
            upderr_log (3,"$inst_spec->{product} already exists on local node, skipping.\n");
            next;			
            }
        # add to install list 
        $inst_list[$ninst++] = $inst_spec;		

        # dump any pending errors
        upderr_get (); 
	upderr_clear();
        }

    my @rev_list = ();
    @rev_list = reverse @dist_list;

    foreach $inst_spec (@rev_list)  {
	 unless (updcmd_split_get_locations($dist_url, $inst_spec,
		    \%has_overlays, \%name_to_spec, \%parent_prod)) {
	     if( !$g_flags{opt_i} ) { return 0; } else { next; }
	}
        # dump any pending errors
        upderr_get (); 
	upderr_clear();
        }


  foreach $inst_spec (@inst_list) {
     upderr_log (0,"informational: beginning install of $inst_spec->{product}.\n");
     upderr_get (); upderr_clear();



     # check for overwrites
     if ( -d $inst_spec->{unwind_prod_dir} && 
		! defined($has_overlays{$inst_spec->{product}})) {
        &upderr_log(0,"error: product " .
			$inst_spec->{product} .
			" directory " .
			$inst_spec->{unwind_prod_dir} .
			" already exists -- will not overwrite ");
        upderr_get (); upderr_clear();
        if( !$g_flags{opt_i} ) { return 0; } else { next; }
     }


     unless (updcmd_split_download_proddir( $dist_url, $inst_spec,
			\%has_overlays, \%name_to_spec, \%parent_prod)) {
         updcmd_split_cleanup($inst_spec,\%has_overlays,\%parent_prod);
         if( !$g_flags{opt_i} ) { return 0; } else { next; }
     }

     unless (updcmd_split_download_ups( $dist_url, $inst_spec,
			\%has_overlays, \%name_to_spec, \%parent_prod )) {
         updcmd_split_cleanup($inst_spec,\%has_overlays,\%parent_prod);
         if( !$g_flags{opt_i} ) { return 0; } else { next; }
     }

    unless (updcmd_split_download_table( $dist_url, $inst_spec,
			\%has_overlays, \%name_to_spec, \%parent_prod )) {
         updcmd_split_cleanup($inst_spec,\%has_overlays,\%parent_prod);
         if( !$g_flags{opt_i} ) { return 0; } else { next; }
    }

     # now we have all the files, we can just declare it...

     # local node shouldn't be distribution node

     delete $inst_spec->{archive_file};	

     unless (&updstb_predeclare( $inst_spec )) {
         updcmd_split_cleanup($inst_spec,\%has_overlays,\%parent_prod);
         if( !$g_flags{opt_i} ) { return 0; } else { next; }
     }

     unless (&updups_declare(0,$inst_spec,{},$g_flags{opt_C}?"-C":"")) {
         updcmd_split_cleanup($inst_spec,\%has_overlays,\%parent_prod);
         if( !$g_flags{opt_i} ) { return 0; } else { next; }
     }

     if ( defined($g_flags{opt_G}) && int(keys(%{$g_flags{opt_G}})) > 0 ) {
        unless (&updups_declare(0,$inst_spec,$g_flags{opt_G},
                $g_flags{opt_C}?"-C":"")) {
             if( $g_flags{opt_i} ) { return 0; } else { next; }
        }
    }

    unless (&updstb_postdeclare( $inst_spec )) {
         if( $g_flags{opt_i} ) { return 0; } else { next; }
    }

    # dump any pending errors
    upderr_get (); 
    upderr_clear();
  } 
  #end of second loop

  foreach $inst_spec (@inst_list) {
      my $inst_note;
      $inst_note = $inst_spec->{unwind_ups_dir} . "/INSTALL_NOTE";
      if ( -f $inst_note ) {
	 upderr_log(0,"informational: product $inst_spec->{product} has an INSTALL_NOTE; \n\tyou should read $inst_note.");
      }
  }

     # get local dependency list and compare against
     # distribution list. if they are different, output 
     # the appropriate ups declare commands to put them in sync.
     # BUT only do this if we are really installing

    unless (defined $g_flags{opt_s}) {
        $savedb = $install_spec->{database};
        delete $install_spec->{database}; 
        &updups_depend(0,$install_spec,\@loc_list,1);
        $install_spec->{database} = $savedb;

        return 0 unless (&updchk_depend(\@loc_list,\@dist_list,\*STDOUT));
        }

     # dump any pending errors
     upderr_get (); 
     upderr_clear();
     return 1;
}

sub override {
     my($inst_spec, $opt, $var) = @_;
     my @list;
     my $unwindvar;
     
     $unwindvar = "unwind_" . $var;
     if ( defined($g_flags{$opt}) ) {
	@list = split(':',$g_flags{$opt});
	$inst_spec->{$unwindvar} = $list[0];
	$inst_spec->{$var} = $list[$#list];
     }
}

sub updcmd_split_cleanup {
     my ($inst_spec,$has_overlays,$parent_prod) = @_;

     upderr_log(1,"Cleaning up...");
     if ( !(defined($has_overlays->{$inst_spec->{product}}) && $has_overlays->{$inst_spec->{product}} ) &&
	    $parent_prod->{$inst_spec->{product}} eq $inst_spec->{product}) {

	if ( -d  $inst_spec->{unwind_prod_dir}) {
	    upderr_log(0,"informational: Starting remove of $inst_spec->{unwind_prod_dir} in background");
	    chdir( "/" );
	    rename($inst_spec->{unwind_prod_dir},$inst_spec->{unwind_prod_dir}.".fail");
	    system("rm -rf $inst_spec->{unwind_prod_dir}.fail &");
        }
	if ( -d  $inst_spec->{unwind_ups_dir}) {
	    upderr_log(0,"informational: starting remove of $inst_spec->{unwind_ups_dir} in background");
	    chdir( "/" );
	    rename($inst_spec->{unwind_ups_dir},"$inst_spec->{unwind_ups_dir}.fail");
	    system("rm -rf $inst_spec->{unwind_ups_dir}.fail &");
        }
     } else {
        upderr_log(1,"Not removing directories due to overlays...");
     }
}


#-------------------------------------------------------------------------
# <return> = updxyz_<functions name>( <input> )
#
# <description>
#
# Input : <input>
# Output: <output>
# Return: <return>
#

sub updcmd_update
{
    my ($install_spec,$component) = @_;
    my $errstr;			# error string
    my @dist_list;			# distribution node dependencies
    my @loc_list;			# local node dependencies
    my $inst_spec;			# instance spec
    my $exist;			# existance flag
    my $ninst = 0;			# number of instances to install
    my @inst_list;			# list of instances to install
    my %has_overlays;		# defined for prodname if it has overlays
    my %name_to_spec;		# hash to find named products fast
    my %parent_prod;		# hash of parent products
    my $dist_url;			# distribution url (a hash reference)
    my $dist_db;
    my $status;			# subroutine return status
    my $savedb;			# temporary slot for database
    my $local_spec;		# local instance spec of product

    #
    &upderr_log(1,"updcmd_update($install_spec,$component)\n");

    return 0 unless 
       updcmd_split_init(\$dist_url,$install_spec,\@dist_list,\$dist_db);

    upderr_log(3,"after init:\n\tdist_node is $dist_url->{node},\n\tdist_list is ".
		  join(":", @dist_list). "\n\tdist_db is $dist_db\n" );

    # for each product in the dependency list
    # ---------------------------------------
    upderr_log(3, "going through depend list $#dist_list $dist_url->{node} \n");
    foreach $inst_spec (@dist_list)  {
        upderr_log (3,"looking at $inst_spec->{product}.\n");
        updcmd_split_bookkeeping( $dist_url, $inst_spec, 
			\%has_overlays, \%name_to_spec, \%parent_prod, \$exist,
			\$local_spec);
        if ( !$exist ) {
            upderr_log (0,"$inst_spec->{product} doesn't exist on local node!\n");
            next;			
            }

        my @datelist1;
        my @datelist2;
        @datelist1 = split(':',$inst_spec->{modified});
        @datelist2 = split(':',$local_spec->{modified});
        upderr_log(2, "remote modified date: |$datelist1[$#datelist1]|\n");
        upderr_log(2, "local modified date: |$datelist2[$#datelist2]|\n");
	if ( $datelist1[$#datelist1] lt $datelist2[$#datelist2]  ) {
	    upderr_log (2, "$inst_spec->{product} is up to date\n");
	    next;
	    }

        # add to install list 
        $inst_list[$ninst++] = $inst_spec;		

        # dump any pending errors
        upderr_get (); 
	upderr_clear();
        }

  foreach $inst_spec (@inst_list) {
     upderr_log (0,"Updating $inst_spec->{product}.\n");

     unless (updcmd_split_get_locations($dist_url, $inst_spec,
			\%has_overlays, \%name_to_spec, \%parent_prod)) {
         if( $g_flags{opt_i} ) { return 0; } else { next; }
     }
     if ($local_spec->{'@prod_dir'} != $inst_spec->{unwind_prod_dir} ) {
	 upderr_log(0, "Notice: prod_dir mismatch: $inst_spec->{product}\n");
	 $inst_spec->{unwind_prod_dir} = $local_spec->{'@prod_dir'};
	 }

     if ( $component =~ m/ups_dir/ ) {

	 if ($local_spec->{'@ups_dir'} != $inst_spec->{unwind_ups_dir} ) {
	     upderr_log(0, "Notice: ups_dir mismatch: $inst_spec->{product}\n");
	     $inst_spec->{unwind_ups_dir} = $local_spec->{'@ups_dir'};
	     }

         return 0 unless 
	     updcmd_split_download_ups( $dist_url, $inst_spec,
			\%has_overlays, \%name_to_spec, \%parent_prod) ||
	     $g_flags{opt_i};

     }
     if ( $component =~ m/table_file/ ) {
        my $table_file_full = 
           "$inst_spec->{unwind_table_dir}/$inst_spec->{table_file}";
        my $local_table_file_full = 
           "$local_spec->{'@table_file'}";

	 if ($table_file_full != $local_table_file_full ) {
	     upderr_log(0, "Notice: table file mismatch: $inst_spec->{product}\n");
	     $table_file_full = $local_table_file_full;
	     }

         if ($g_flags{opt_s}) {
	     upderr_log(0,"I would rename $table_file_full to .bak now");
         } else {
	     rename($table_file_full, $table_file_full . ".bak")
         }

         if (! updcmd_split_download_table( $dist_url, $inst_spec,
			\%has_overlays, \%name_to_spec, \%parent_prod) ) {
	     rename( $table_file_full . ".bak", $table_file_full );
	     return 0 unless ($g_flags{opt_i});
	 }
     }
     # dump any pending errors
     upderr_get (); 
     upderr_clear();
     updups_touch(0,$local_spec);
     } 
     #end of second loop

     return 1;
}

# --------------------------------------------------------------------
# routines split out to share between updcmd_install and updcmd_update 
#

sub updcmd_split_init {
    my ($dist_url, $install_spec, $dist_list, $dist_db ) = @_;
    return 0 unless (&updchk_instance_spec($install_spec,"unique"));

    return 0 unless ( ($$dist_url = &upduti_parse_url( $g_flags{opt_h} )) );

    if (defined $g_flags{opt_z}) { 
       $$dist_db = $g_flags{opt_z}; 
       $install_spec->{database} = $$dist_db;
       }
    else { 
       $$dist_db = "" ;
       delete $install_spec->{database};
       }

    # build the list of remote dependencies, if -j is specified, the 
    # remote dependency list is the same 

    if ( defined($g_flags{opt_j}) ) { 
       $${dist_list}[0] = $install_spec; 
       return 0 unless (updups_find($g_flags{opt_h}, $install_spec, 
			"_hack1", "database"));
       }
    else { 
       return 0 unless (&updups_depend($$dist_url->{str},$install_spec,$dist_list)); 
       }

    if ( $#$dist_list == -1 ) {
       upderr_log(0,"Error: Distribution dependency list empty (!)".
	 " for product $install_spec->{product}.\n");
       return 0;
       }
    return 1;
}

sub updcmd_split_bookkeeping {
    my ($dist_url, $inst_spec, $has_overlays, 
        $name_to_spec, $parent_prod, $exist , $ret_local_spec) = @_;
    my $savedb;			# saved database
    my $status;
    my %prop_info;
    my $prop_archive_file;	# proprietary info
    my @externalfiles;		# files external to product
    my $local_spec;		# inst spec for install of product (if any)
    my $proprietarylist;        # list of proprietary products

    upderr_log(3,"args($dist_url->{node}, $inst_spec, $has_overlays, $name_to_spec, $parent_prod, $exist)");
    # we don't want the chain name messing with our future declares ...
    #
    if (defined $inst_spec->{chain}) { 
	delete $inst_spec->{chain}; 
    }

    $proprietarylist = updcmd_get_proprietarylist($dist_url,$inst_spec->{database});

    $name_to_spec->{$inst_spec->{'product'}} = $inst_spec;


    # already exists on local node, skip.
    # we want to look in $PRODUCTS, so temporarily reset the 
    # distribution database.
    # ------------------------------------------------------
    $local_spec = {%$inst_spec};
    delete $local_spec->{database}; 
    return 0 unless ( updups_exist(0,$local_spec,$exist));
    if ( $$exist ) {
       updups_find(0, $local_spec,
			    '@prod_dir',
			    '@table_file',
			    '@ups_dir',
			    "modified",
		  );
	$$ret_local_spec = $local_spec;
    }
    delete $inst_spec->{prod_dir};
    return 0 unless 
       updups_find($dist_url->{str},$inst_spec,
			    "_upd_overlay",
			    "database",
			    "archive_file",
			    "prod_dir",
			    '@prod_dir',
			    "table_dir",
			    "table_file",
			    '@table_file',
			    "ups_dir",
			    "modified",
			    "description",
			    "origin",
		 );

    # -- don't really  need this for update, but it won't hurt...

    # if it's an overlay, we need some bookkeeping
    # --------------------------------------------
    if ( defined($inst_spec->{'_upd_overlay'}) && 
	      $inst_spec->{'_upd_overlay'} ne '' &&
	      $inst_spec->{'_upd_overlay'} ne '""' ) {

       $has_overlays->{ $inst_spec->{product} } = 1;
       $has_overlays->{ $inst_spec->{'_upd_overlay'} } = 1;
       $parent_prod->{$inst_spec->{product}} = 
	    $inst_spec->{'_upd_overlay'};
       upderr_log (3,"$inst_spec->{product} is overlaid on $inst_spec->{'_upd_overlay'}.\n");
       } 
    else {
       $parent_prod->{$inst_spec->{product}} = 
	    $inst_spec->{'product'};
       }

    # it's proprietary, get proprietary information
    # ---------------------------------------------

    if ( $proprietarylist =~ m/:$inst_spec->{product}:/ ) {
        upderr_log (3,"$inst_spec->{product} is proprietary.\n");
        $status = updui_prompt_propriety ($inst_spec, \%prop_info, \$prop_archive_file);
        unless ( $status ) {			# error
           if ( defined($g_flags{opt_i}) )		# we want to ignore errors
	      {
	      &upderr_log (0,"Skipping $inst_spec->{product}.\n");
	      &upderr_get (); upderr_clear();
	      return 0;				# check next instance
	      }
           else { 
	      return 0; 				# exit on error
	      }
          }

	 # no longer needed with new proprietary structure -- mengel
	 #
         # $inst_spec->{'archive_file'} = $prop_archive_file;

         $inst_spec->{'propriety'} = { %prop_info }; 

 	 upderr_log(3, "propriety->group_name is $inst_spec->{propriety}->{group_name}\n");
 	 upderr_log(3, "propriety->group_pass is $inst_spec->{propriety}->{group_pass}\n");
         }
     return 0 unless 
	 updups_get( $dist_url->{str}, $inst_spec, \@externalfiles );

    if ( $#externalfiles > -1 ) {
        upderr_log(0,"Warning: for product " .
		    $inst_spec->{'product'} . 
		    ", the files: \n   " . 
		    join("\n   ",@externalfiles).
		    "\nare external, and need to be fetched manually\n");
        upderr_get (); 
	upderr_clear();
        }
	return 1;
}

sub updcmd_split_get_locations {
    my ($dist_url, $inst_spec, $has_overlays, 
	$name_to_spec, $parent_prod ) = @_;

    # make sure parent spec exists!
    my $parent_spec = $name_to_spec->{$parent_prod->{$inst_spec->{product}}};
    unless ($parent_spec)
       {
       upderr_log(0,"ERROR: $inst_spec->{product}'s parent product ".
                    "\"$parent_prod->{$inst_spec->{product}}\" ".
                    "not in dependency list!\n");
       return 0;
       }

    return 0 unless  
	(&updstb_get_db( $inst_spec, $parent_spec ));
   
    # Used to say:
    # get unwind dir first, because getdb will *change* the database
    # but this is wrong, callouts should use dist_db, we need the db
    # so we can find ${PROD_DIR_PREFIX}...

    return 0 unless 
	(&updstb_get_proddir( $inst_spec, $parent_spec ));
     override( $inst_spec, 'opt_r', 'prod_dir' );
	
    return 0 unless
	(&updstb_get_upsdir( $inst_spec ));
     override( $inst_spec, 'opt_U', 'ups_dir' );

    return 0 unless
	(&updstb_get_tablefile( $inst_spec ));
     override( $inst_spec, 'opt_M', 'table_dir' );
     override( $inst_spec, 'opt_m', 'table_file');

     # now we should have all we need to download and declare the product.

     if ( !defined($inst_spec->{dist_archive_file} ) ||
	  ($inst_spec->{dist_archive_file} eq '""')  ||
	  ($inst_spec->{dist_archive_file} eq '') ) {
	# we have a live database, so we use the ftp server to 
	# tar it up on the fly...

	$inst_spec->{dist_archive_file} = "ftp://" . $dist_url->{node} . 
				     $inst_spec->{'@prod_dir'} . "/..tar";
        }

     # note we strip '.tar' or .zip off the end, 'cause updxfr_dir adds it
     # back on (and may at some point tack on a .Z or .gz for compression...)

     return 1;
}

sub updcmd_split_download_proddir {
    my($dist_url, $inst_spec, $has_overlays, $name_to_spec, $parent_prod) = @_;
    my ($proto,$host,$port,$dir,$file,$arch); # fields split from archive_file URL
     # fetch & unwind main product.
     ($proto,$host,$port,$dir,$file,$arch) = 
		upduti_split_archive_file($inst_spec->{dist_archive_file});

     if ($proto ne "ftp") {
        &upderr_log (0,"$inst_spec->{product} specifes $proto transfer, only ftp is supported.\n");
        upderr_get (); upderr_clear();
	if ( !$g_flags{opt_i} ) {
	    return 0;
	}
     }

     my $k;
     foreach $k ( keys(%{$inst_spec->{propriety}})) {
       upderr_log(4,"propriety{$k} = $inst_spec->{propriety}->{$k}\n");
     }

     return 0 unless
        &updxfr_dir( $host , $dir, $file, $inst_spec->{unwind_prod_dir},
		$inst_spec->{propriety}, $arch );


     unless ( &updchk_manifest( $inst_spec->{unwind_prod_dir},$inst_spec->{product} )) {
        upderr_get (); upderr_clear();
	return 0 unless $g_flags{opt_i};
        }

    return 1;
}

sub updcmd_split_download_ups {
   my ($dist_url, $inst_spec, $has_overlays, $name_to_spec, $parent_prod) = @_;
     # get ups dir -- if there is one
     # this may stomp it right overtop of where it already was,
     #  but it's too hard to figure out when we shouldn't...

    my ($proto,$host,$port,$dir,$file,$arch);# fields split from archive_file URL
     ($proto,$host,$port,$dir,$file,$arch) = 
		upduti_split_archive_file($inst_spec->{dist_archive_file});
	 
     if ($proto ne "ftp") {
        &upderr_log (0,"$inst_spec->{product} specifes $proto transfer, only ftp is supported.\n");
        upderr_get (); upderr_clear();
	if ( !$g_flags{opt_i} ) {
	    return 0;
	}
     }

     if ( $inst_spec->{dist_ups_dir} ) 
        {
        if (! ($inst_spec->{dist_ups_dir} =~ m|^/|) ) {
	    $inst_spec->{dist_ups_dir} = $inst_spec->{'@prod_dir'} . "/" .
					$inst_spec->{dist_ups_dir};
           }
	if ( $g_flags{opt_s} ) {
	  upderr_log( 0,"I would: chdir to $inst_spec->{unwind_prod_dir}\n");
	}
	elsif (! chdir($inst_spec->{unwind_prod_dir})) {
	  upderr_log(0,"error: Could not chdir to $inst_spec->{unwind_prod_dir}\n");
	  return 0;
	}

        if ( -d $inst_spec->{unwind_ups_dir} ) {
	    #
            # best effort to make files writable so we can update them...
	    #
	    system("chmod -R +w $inst_spec->{unwind_ups_dir} > /dev/null 2>&1");
        }

	return 0 unless
	    &updxfr_dir( $host, 
		      $inst_spec->{dist_ups_dir}, 
		      ".", 
		      $inst_spec->{unwind_ups_dir},
		      $inst_spec->{propriety}, 
		      "tar"
	            );
     	}

# there is no distribution node ups dir. let's delete this info
# so we won't confuse the ups declare command

     else
        {
        delete $inst_spec->{ups_dir};
        }

return 1;
}

sub updcmd_split_download_table {
   my ($dist_url, $inst_spec, $has_overlays, $name_to_spec, $parent_prod) = @_;
     # get ups dir -- if there is one
     # this may stomp it right overtop of where it already was,
     #  but it's too hard to figure out when we shouldn't...

    my ($proto,$host,$port,$dir,$file,$arch);	# fields split from archive_file URL
     ($proto,$host,$port,$dir,$file,$arch) = 
		upduti_split_archive_file($inst_spec->{dist_archive_file});
	 
     if ($proto ne "ftp") {
        &upderr_log (0,"$inst_spec->{product} specifes $proto transfer, only ftp is supported.\n");
        upderr_get (); upderr_clear();
	if ( !$g_flags{opt_i} ) {
	    return 0;
	}
     }


     # get table file
     # (this may stomp it right overtop of where it already was,
     #  but it's too hard to figure out when we shouldn't...)

# --if updating table file

     if ($inst_spec->{'@table_file'})
        {
        my $table_file_full = 
           "$inst_spec->{unwind_table_dir}/$inst_spec->{table_file}";
	if ( $g_flags{opt_s} ) {
	  upderr_log( 0,"I would: chdir to $inst_spec->{unwind_prod_dir}\n");
	}
        elsif (! chdir($inst_spec->{unwind_prod_dir})) 
           {
           upderr_log(0,"error: Could not chdir to $inst_spec->{unwind_prod_dir}\n");
           return 0;
           }
        return 0 unless 
           (&updxfr_file( $host,"/",$inst_spec->{'@table_file'}, 
               $table_file_full . ".new" )); 
        unlink($table_file_full) if ( -f $table_file_full );
        rename( $table_file_full . ".new", $table_file_full ); 
        }

# there is no distribution node table file. let's delete this info
# so we won't confuse the ups declare command

     else
        {
        delete $inst_spec->{table_file};
        delete $inst_spec->{table_dir};
        }

return 1;
}
#-------------------------------------------------------------------------
# <status> = upd_fetch ( $inst_spec)
#
# <description>
#
# Input : hash $inst_spec
# Output: <output>
# Return: <return>
#

sub updcmd_fetch
{
    my ( $inst_spec ) = @_;

    my $status = 0;
    my $local = "";
    my %prop_info;
    my $prop_archive_file;

    my $url = undef;
    my $proprietarylist;
    return 0 unless ( ($url = upduti_parse_url( $g_flags{opt_h} )) );

    upderr_log(1,"args $inst_spec\n");
    $proprietarylist = updcmd_get_proprietarylist($url,$inst_spec->{database});

    if ( $proprietarylist =~ m/:$inst_spec->{product}:/ ) {
        upderr_log (3,"$inst_spec->{product} is proprietary.\n");
        $status = updui_prompt_propriety ($inst_spec, \%prop_info, \$prop_archive_file);
    }
	
    if ( defined ( $g_flags{opt_J} ) ) { 		# have a filename
	    my @findlist;
	    upderr_log(2, "Have -J Flag: $g_flags{opt_J}\n");

	    @findlist = ("prod_dir", "qualifiers", "database", "ups_dir");
	    if ($g_flags{opt_J} =~ m/\A@.*/) {
		push(@findlist,$g_flags{opt_J});
	    }
											    # fill in the rest of inst_spec
	    $status = updups_find($url->{str}, $inst_spec, @findlist);
	    if (! $status) {
		    upderr_log(2, "updups_find failed...Status: $status\n");
		    return ($status);
	    }

	    if ($g_flags{opt_J} =~ m/\A@.*/) {
		if (!defined($inst_spec->{$g_flags{opt_J}})) {
		    upderr_log(0,"Error: Unable to expand $g_flags{opt_J}");
		    return 0;
		}
		$g_flags{opt_J} = $inst_spec->{$g_flags{opt_J}};
	    }
											    # now get the file
	    $local = $g_flags{opt_J};
	    $local =~ s/.*[\\\/]//;
	    $local =~ "./" . $local;
	    if ( -f $local ) {
		unlink $local;
	    }
	    $status = updxfr_file( $url->{node}, $inst_spec->{prod_dir},
			    $g_flags{opt_J}, $local );
			    
	    if (! $status) {
		    upderr_log(2, "Can't transfer file: $g_flags{opt_J}\n");
		    return ($status);
	    }
            return 1;
	} else { # no filename then do a ls
	    upderr_log(3, "-J filename flag not defined. Doing a dir ls\n");
	    $status = updxfr_ls($url->{node}, $inst_spec->{prod_dir});
	    return $status;
	}
        upderr_log(0,"error: upd fetch called with bad arguments\n");
	return 0;
}

sub updcmd_delproduct {
    my ($inst_spec ) = @_;
    my ($dist_url,		# url of regular ups.cgi script
	$decl_url,		# url of ups-decl.cgi script
	$res,			# result of running command
        $existing,		# product exists
	);			
    upderr_log(3, "entering updcmd_delproduct" );



    return 0 unless (&updchk_instance_spec($inst_spec,"unique"));
    return 0 unless (($dist_url = &upduti_parse_url( $g_flags{opt_h} )) );
    return 0 unless updups_exist( $decl_url->{str}, $inst_spec, \$existing );
    if ($inst_spec->{database} eq '' ) {
        delete $inst_spec->{database}
    }
    updups_find($dist_url->{str}, $inst_spec, 'prod_dir','database');

    # check arguments for validity
    if ( $dist_url->{path} eq '' ) {
        $dist_url->{str} .= '/cgi-bin/ups.cgi';
        $dist_url->{path} = '/cgi-bin/ups.cgi';
    }
    $decl_url = {%$dist_url};
    $decl_url->{str} =~ s/ups.cgi/ups-decl.cgi/;
    $decl_url->{path} =~ s/ups.cgi/ups-decl.cgi/;


    if ($inst_spec->{chain} eq '' ) {
        delete $inst_spec->{chain}
    }

    #$res = updups_undeclare( $decl_url->{str}, $inst_spec , $inst_spec , "-Y -C" );
    $res = updups_undeclare( $decl_url->{str}, $inst_spec , $inst_spec , "-Y" );
    return $res;
}

sub updcmd_chainproduct {
    my ($inst_spec ) = @_;
    my ($dist_url,		# url of regular ups.cgi script
	$decl_url,		# url of ups-decl.cgi script
	$res,			# result of running command
	$old_spec,		# spec for previous chained version
        $existing,		# product exists
        $chain,			# product chain to save
	);			
    upderr_log(0, "notice: chainproduct is replaced by modproduct." );
    return 0;
}

sub updcmd_add_mod_product {
   
    my ($inst_spec, $modflag ) = @_;
    my ($puttarfile, 		# tarfile name on server
	$puttblfile,  		# table file name on server
	@movedto, 		# file locations returned by cgi scripts
	@errs,			# errors returned by cgi scripts
    	$move_url,		# url of upd.cgi script
	$dist_url,		# url of regular ups.cgi script
	$decl_url,		# url of ups-decl.cgi script
	$dist_db,		# database on dist. node
	$cmd,			# command string
	$existing,		# product exists on server already
	$addhost,		# host on which we are adding
	$res,			# result of running command
	$chain,			# chain from command line
        $made_tarfile		# flag if we need to clean up a tarfile
	);			

    $made_tarfile = 0;
    $res = 1;
    upderr_log(3, "entering updcmd_add_mod_product modflag == $modflag" );

    # check arguments for validity

    return 0 unless (&updchk_instance_spec($inst_spec,"unique"));
    return 0 unless (($dist_url = &upduti_parse_url( $g_flags{opt_h} )) );

    #
    # take chain out of the inst_spec until we have done checks for info
    # and existance...
    #
    $chain = $inst_spec->{chain};
    delete $inst_spec->{chain};

    return 0 unless updups_exist( $dist_url->{str}, $inst_spec, \$existing );


    # find urls for move and declare ups scripts

    if ( $dist_url->{path} eq '' ) {
        $dist_url->{str} .= '/cgi-bin/ups.cgi';
        $dist_url->{path} = '/cgi-bin/ups.cgi';
    }
    $move_url = {%$dist_url};
    $decl_url = {%$dist_url};
    $move_url->{str} =~ s/ups.cgi/upd.cgi/;
    $decl_url->{str} =~ s/ups.cgi/ups-decl.cgi/;
    $move_url->{path} =~ s/ups.cgi/upd.cgi/;
    $decl_url->{path} =~ s/ups.cgi/ups-decl.cgi/;

    $addhost = $dist_url->{node};

    # determine database on distrib host

    if (defined $g_flags{opt_z}) { 
        $dist_db = $g_flags{opt_z}; 
        $inst_spec->{database} = $dist_db;
    } else { 
       $dist_db = "" ;
       delete $inst_spec->{database};
    }

    if ( !$modflag ) {
        #
        # special cases for addprouct
        #  - default ups,table,prod_dir from local declaration, if any
        #  - make archive file if we have a prod_dir but no archive_file
        #
 
        if ($existing) {
	    upderr_log(0,"error: addproduct cannot overwrite an existing instance.");
 	    return 0;
	}

        updups_find(0,$inst_spec,
	  'archive_file', '@ups_dir', '@table_file', 'table_file', '@prod_dir');

        $inst_spec->{ups_dir} = $inst_spec->{'@ups_dir'} 
		unless defined $inst_spec->{ups_dir};

	unless (defined $inst_spec->{table_dir}) {
	    $inst_spec->{table_dir} = $inst_spec->{'@table_file'};
	    $inst_spec->{table_dir} =~ s|\A(.*)/[^/]*\Z|$1|;
	}
	unless (defined $inst_spec->{table_file}) {
	    $inst_spec->{table_file} = $inst_spec->{'@table_file'};
	    $inst_spec->{table_file} =~ s|\A.*/([^/]*)\Z|$1|;
	}

        if ((!defined ($inst_spec->{archive_file}) || 
			$inst_spec->{archive_file} eq '') && 
				defined($inst_spec->{'@prod_dir'})) {

           $inst_spec->{archive_file} = 
		    upduti_make_tarfile($inst_spec->{'@prod_dir'},$inst_spec->{product});
	   if ( $inst_spec->{archive_file} eq '') {
		upderr_log("Unable to make tarfile of product root.");
		return 0;
 	   }
           $made_tarfile = 1;
        }
        if ((!defined ($inst_spec->{archive_file}) || 
			$inst_spec->{archive_file} eq '') && 
				defined($inst_spec->{'prod_dir'})) {

           $inst_spec->{archive_file} = 
		    upduti_make_tarfile($inst_spec->{'prod_dir'},$inst_spec->{product});
	   if ( $inst_spec->{archive_file} eq '') {
		upderr_log("Unable to make tarfile of product root.");
		return 0;
 	   }
           $made_tarfile = 1;
        }
    } else {
        if (!$existing) {
	    upderr_log(0,"error: modproduct can only affect an existing instance.");
 	    return 0;
	}
    }

    $inst_spec->{chain} = $chain;

    delete $inst_spec->{archive_file} if $inst_spec->{archive_file} eq '';
    delete $inst_spec->{table_file}   if $inst_spec->{table_file} eq '';
    delete $inst_spec->{table_dir}    if $inst_spec->{table_dir} eq '';
    delete $inst_spec->{chain}        if $inst_spec->{chain} eq '';
    delete $inst_spec->{database}     if $inst_spec->{database} eq '';

    upduti_print_instance_spec(3, $inst_spec);

    if (defined $inst_spec->{table_file} &&  !updups_check_table(0,$inst_spec,{})) {
	upderr_log( 0, "upd: specified table file not compatable with options specified");
	return 0;
    }

    # check for mandatory arguments

    if (!defined $inst_spec->{archive_file}  &&
		!defined $inst_spec->{table_file} &&
		!defined $inst_spec->{ups_dir} && 
	        !defined $inst_spec->{chain}    ) {
	upderr_log(0,"error: must specify at least one of -g, -m, -U, or -T");
	return 0;
    }
    upderr_get(); upderr_clear();

    # now upload and install tarfile and/or tablefile

    if (defined $inst_spec->{archive_file} ) {
	if ($existing) {
	    upderr_log(0, "error: unable to update tarfile with modproduct, delete it first");
	    return 0;
	}

	if ($g_flags{opt_s}) {
            upderr_log(0,"I would upload $inst_spec->{archive_file} to $dist_url->{node} now");
            $inst_spec->{archive_file} = "uploaded_".$inst_spec->{archive_file};
	} else {

	# upload the tarfile to /incoming

	$res = updxfr_put_unique($dist_url->{node},"/incoming",\$puttarfile,
					    $inst_spec->{archive_file}) ;


        upderr_get(); upderr_clear();
	# tell upd.cgi to move it someplace useful
	# note that this runs a "upd move_archive_file" on the server node.

	$inst_spec->{archive_file} = "incoming/" . $puttarfile;

	$cmd = "move_archive_file" .
			" -O '" . 
			$g_flags{opt_O} . 
			"' " .
			ispec2str($inst_spec, 'Tfqgz');

        @movedto = ();
        @errs = ();

        
	$res = (0 == updupr_oxc($move_url->{str}, $cmd, \@movedto, \@errs))
	   if ($res);

        upderr_get(); upderr_clear();

        # update our inst_spec with returned locations

        $inst_spec->{database} = $movedto[$#movedto - 2];

	if ($made_tarfile) {
	   unlink($inst_spec->{orig_archive_file});
	}

        if ( $movedto[$#movedto - 1] ne "" ) {
	    $inst_spec->{archive_file} = "ftp://" . $addhost . $movedto[$#movedto - 1];
	} else {
	    delete $inst_spec->{archvive_file};
	}

	$inst_spec->{prod_dir} = $movedto[$#movedto];

	if ($#errs > -1) {
	   upderr_log(0,"output of move_archive_file: " . join( "\n\t", @errs ). "\n");
	}

	}
    } elsif ($existing)  {
	updups_find($dist_url->{str}, $inst_spec, 'prod_dir', 'archive_file' );
    }
    upderr_get(); upderr_clear();

    if (defined $inst_spec->{table_file} ) {
	my $local = $inst_spec->{table_file};

	if ( defined $inst_spec->{table_dir} ) {
	    $local =  $inst_spec->{table_dir} . '/' . $local;
	}

      if ($g_flags{opt_s}) {
	upderr_log(0,"I would upload $local to $dist_url->{node} now");
	$inst_spec->{table_dir} = "uploaded_".$inst_spec->{table_dir};
      } else {

        $res = updxfr_put_unique($dist_url->{node},"/incoming",\$puttblfile, $local)
		if ($res);

        upderr_get(); upderr_clear();

	$inst_spec->{table_file} = $puttblfile;
	$inst_spec->{table_dir} = 'incoming';

        upduti_print_instance_spec(3,$inst_spec);

        $cmd = "move_table_file " . 
			" -O '" . 
 			($modflag ? "modify":"add") .
			$g_flags{opt_O} . 
			"' " .
			ispec2str($inst_spec, 'TmMfqz');


        @movedto = ();
        @errs = ();
        $res = (0 == updupr_oxc($move_url->{str}, $cmd, \@movedto, \@errs))
		if ($res);

        $inst_spec->{database} = $movedto[$#movedto - 2];
        $inst_spec->{table_file} = $movedto[$#movedto - 1];
        $inst_spec->{table_dir} = $movedto[$#movedto];


	if ($#errs > -1) {
	   upderr_log(0,"output of move_table_file: " . join( "\n\t", @errs ) . "\n");
	}

        upderr_get(); upderr_clear();
      }

    } elsif ($existing) {
	updups_find($dist_url->{str}, $inst_spec, 'table_dir', 'table_file' );
    }

    if (defined $inst_spec->{ups_dir} && $inst_spec->{ups_dir} ne '' ) {
        my $local = upduti_make_tarfile( $inst_spec->{ups_dir},$inst_spec->{product} );

      if ($g_flags{opt_s}) {
	upderr_log(0,"I would upload $local to $dist_url->{node} now");
	$inst_spec->{ups_dir} = "uploaded_".$inst_spec->{ups_dir};
      } else {

        $res = updxfr_put_unique($dist_url->{node},"/incoming",\$puttarfile, $local)
		if ($res);

	unlink($local);

        $inst_spec->{ups_dir} = 'incoming/' . $puttarfile;

	$cmd = "move_ups_dir" .
			" -O '" . $g_flags{opt_O} . "' " .
			ispec2str($inst_spec, 'Ufqgz');

        @movedto = ();
        @errs = ();
	$res = (0 == updupr_oxc($move_url->{str}, $cmd, \@movedto, \@errs))
		if ($res);

        upderr_get(); upderr_clear();
        $inst_spec->{database} = $movedto[$#movedto - 1];
        $inst_spec->{ups_dir} = $movedto[$#movedto];

	if ($#errs > -1) {
	   upderr_log(0,"output of move_ups_dir: " . join( "\n\t", @errs ) . "\n");
	}
      }
    }

    upderr_get(); upderr_clear();

    if ( defined $inst_spec->{chain} ) {
        my $old_spec;

	# find existing chain on this product, if any, and remove the chain
	$old_spec = {%$inst_spec};
	delete $old_spec->{chain};
	updups_find($dist_url->{str},$old_spec,'chain');
	if ( defined $old_spec->{chain} && '' ne $old_spec->{chain} ) {
	     delete $old_spec->{version};
	     updups_undeclare( $decl_url->{str}, $old_spec , $old_spec , "-C" );
	}
    }

    if ($existing && (!defined ($inst_spec->{chain}) || $inst_spec->{chain} != ':')) {
        $res = updups_touch( $decl_url->{str}, $inst_spec )
	    if($res);
    } else {
        $res = updups_declare( $decl_url->{str}, $inst_spec , $inst_spec , "-C" )
	    if($res);
    }

    upderr_get(); upderr_clear();

    return $res;
}

sub updcmd_cloneproduct {
    my ($inst_spec) = @_;
    my ($puttarfile, 		# tarfile name on server
	$puttblfile,  		# table file name on server
	@movedto, 		# file locations returned by cgi scripts
	@errs,			# errors returned by cgi scripts
    	$move_url,		# url of upd.cgi script
	$dist_url,		# url of regular ups.cgi script
	$decl_url,		# url of ups-decl.cgi script
	$dist_db,		# database on dist. node
	$cmd,			# command string
	$existing,		# product exists on server already
	$addhost,		# host on which we are adding
	$res,			# result of running command
	$chain,			# chain from command line
        $made_tarfile		# flag if we need to clean up a tarfile
	);			

    $made_tarfile = 0;
    $res = 1;
    upderr_log(3, "entering updcmd_cloneproduct" );

    # check arguments for validity

    return 0 unless (&updchk_instance_spec($inst_spec,"unique"));
    return 0 unless (($dist_url = &upduti_parse_url( $g_flags{opt_h} )) );

    #
    # take chain out of the inst_spec until we have done checks for info
    # and existance...
    #
    $chain = $inst_spec->{chain};
    delete $inst_spec->{chain};

    return 0 unless updups_exist( $dist_url->{str}, $inst_spec, \$existing );


    # find urls for move and declare ups scripts

    if ( $dist_url->{path} eq '' ) {
        $dist_url->{str} .= '/cgi-bin/ups.cgi';
        $dist_url->{path} = '/cgi-bin/ups.cgi';
    }
    $move_url = {%$dist_url};
    $decl_url = {%$dist_url};
    $move_url->{str} =~ s/ups.cgi/upd.cgi/;
    $decl_url->{str} =~ s/ups.cgi/ups-decl.cgi/;
    $move_url->{path} =~ s/ups.cgi/upd.cgi/;
    $decl_url->{path} =~ s/ups.cgi/ups-decl.cgi/;

    $addhost = $dist_url->{node};

    # determine database on distrib host

    if (defined $g_flags{opt_z}) { 
        $dist_db = $g_flags{opt_z}; 
        $inst_spec->{database} = $dist_db;
    } else { 
       $dist_db = "" ;
       delete $inst_spec->{database};
    }

    if (!$existing) {
	upderr_log(0,"error: product must exist");
	return 0;
    }

    # -----------------------------------------
    # find info of existing instance
    return 0 unless updups_find( $dist_url->{str}, $inst_spec, 
					'@prod_dir',
					'@ups_dir',
					'@table_file',
					'archive_file',
				);

    $inst_spec->{prod_dir} = $inst_spec->{'@prod_dir'};
    $inst_spec->{ups_dir} = $inst_spec->{'@ups_dir'};
    $inst_spec->{'@table_file'} =~ m|(.*)/(.*)|;
    $inst_spec->{table_dir} = $1;
    $inst_spec->{table_file} = $2;

    $res = 1;

    # -----------------------------------------
    # override flags from -G info

    my $x_args;
    $x_args = $g_flags{opt_G};

    foreach ( keys %{$x_args} ) {
	$inst_spec->{$_} = $x_args->{$_} if ($x_args->{$_} ne '');
    }

    # -----------------------------------------
    # clone archive file
    if (defined( $inst_spec->{archive_file} ) && $inst_spec->{archive_file} ne '' ) {
	if ($g_flags{opt_s}) {
            upderr_log(0,"I would clone $inst_spec->{archive_file} on $dist_url->{node} now");
            $inst_spec->{archive_file} = "cloned".$inst_spec->{archive_file};
	} else {

        $inst_spec->{archive_file} =~ s|\Aftp://[A-Za-z0-9_.:]+||;

	$cmd = "move_archive_file " . " -O '" . "clone" .  $g_flags{opt_O} . "' " .
			ispec2str($inst_spec, 'Tfqz');
	@movedto = ();
	@errs = ();
	$res = (0 == updupr_oxc($move_url->{str}, $cmd, \@movedto, \@errs))
		if ($res);
	$inst_spec->{archive_file} = "ftp://" . $addhost . $movedto[$#movedto - 1];
	$inst_spec->{prod_dir} = $movedto[$#movedto];
        }
    }

    # -----------------------------------------
    # clone ups dir
    if (defined( $inst_spec->{ups_dir} ) && $inst_spec->{ups_dir} ne '' ) {
	if ($g_flags{opt_s}) {
            upderr_log(0,"I would clone $inst_spec->{ups_dir} on $dist_url->{node} now");
            $inst_spec->{archive_file} = "cloned".$inst_spec->{ups_dir};
	} else {

	$cmd = "move_ups_dir " . " -O '" . "clone" .  $g_flags{opt_O} . "' " .
			ispec2str($inst_spec, 'Ufqz');
	@movedto = ();
	@errs = ();
	$res = (0 == updupr_oxc($move_url->{str}, $cmd, \@movedto, \@errs))
		if ($res);
        $inst_spec->{database} = $movedto[$#movedto - 1];
        $inst_spec->{ups_dir} = $movedto[$#movedto];
        }
    }

    # -----------------------------------------
    # clone table_file
    if (defined( $inst_spec->{table_file} ) && $inst_spec->{table_file} ne '' ) {
	if ($g_flags{opt_s}) {
            upderr_log(0,"I would clone $inst_spec->{'@table_file'} on $dist_url->{node} now");
            $inst_spec->{table_dir} = "cloned".$inst_spec->{table_dir};
	} else {

	$cmd = "move_table_file " . " -O '" . "separate_tables:clone" .  $g_flags{opt_O} . "' " .
			ispec2str($inst_spec, 'Mmfqz');
	@movedto = ();
	@errs = ();
	$res = (0 == updupr_oxc($move_url->{str}, $cmd, \@movedto, \@errs))
		if ($res);
        $inst_spec->{database} = $movedto[$#movedto - 2];
        $inst_spec->{table_file} = $movedto[$#movedto - 1];
        $inst_spec->{table_dir} = $movedto[$#movedto];
        }
    }


    # -----------------------------------------
    # declare new instance
    $res = updups_declare( $decl_url->{str}, $inst_spec , $inst_spec , "-C" )
	if($res);
}

sub updcmd_get_move_locations {
    my ($inst_spec) = @_;
    my ($exists);

    # check if product exists locally

    updups_exist(0,$inst_spec,\$exists);

    # look up in user callouts...

    return 0 unless 
	(&updstb_get_db( $inst_spec, $inst_spec ));
    return 0 unless 
	(&updstb_get_proddir( $inst_spec, $inst_spec ));
    return 0 unless
	(&updstb_get_tablefile($inst_spec));
    return 0 unless
	(&updstb_get_upsdir($inst_spec));

    if ($exists) {
        
        # if product exists already, look up and override stuff
	# thats defined
	#
        delete $inst_spec->{table_dir};
        delete $inst_spec->{table_file};
        delete $inst_spec->{'@table_file'};
        delete $inst_spec->{ups_dir};
        delete $inst_spec->{'@ups_dir'};
	updups_find(0,$inst_spec,
		        'database', 'table_dir', 
			'table_file','@table_file',
			'prod_dir', '@prod_dir',
			'ups_dir', '@ups_dir');

        unless ($inst_spec->{'@prod_dir'} eq '') {
	   $inst_spec->{unwind_prod_dir} = $inst_spec->{'@prod_dir'}
	}

        unless ($inst_spec->{'@table_file'} eq '') {
	    $inst_spec->{unwind_table_dir} = $inst_spec->{'@table_file'};
	    $inst_spec->{unwind_table_dir} =~ s|(.*)/.*|$1|;
	}
        unless ($inst_spec->{'@ups_dir'} eq '') {
	    $inst_spec->{unwind_ups_dir} = $inst_spec->{'@ups_dir'};
	}
    }
    upduti_print_instance_spec(3,$inst_spec); 
    return 1;
}

sub updcmd_move_archive_file {
    my ($inst_spec) = @_;	 # instance spec for product
    my ($suffix, $dest, $ddest); # distrib file suffix, final location
    my $src;
    my $cmd;			 # unwind command
    my $unpack = 		 # stuff to unpack if doing partial unwind
	"ups README man ./ups ./README ./man 'ups/*'";  

    upderr_log(3, "entering updcmd_move_archive_file\n" );

    if (defined($inst_spec->{archive_file})) {
        upderr_log(3, "archive_file is $inst_spec->{archive_file}\n" );
    } else {
        upderr_log(0, "archive_file is not defined!\n" );
	return 0;
    }


    upderr_log(3, "making db and proddir callouts...\n" );

    return 0 unless (updcmd_get_move_locations($inst_spec));

    $suffix = (upduti_split_archive_file("file://localhost/$inst_spec->{archive_file}"))[5];

    if ( defined( $inst_spec->{unwind_archive_file})) {
	$dest = $inst_spec->{unwind_archive_file};
        $ddest = $inst_spec->{ups_archive_file};
    } else {
	$dest = $inst_spec->{unwind_prod_dir}.".$suffix";
	$ddest = $inst_spec->{prod_dir}.".$suffix";
    }

    if ( -d $inst_spec->{unwind_prod_dir} ) {
	upderr_log(0,"Notice: removing previous copy of product directory\n");
        rename( $inst_spec->{unwind_prod_dir},
		"$inst_spec->{unwind_prod_dir}.old");
        system( "rm -rf $inst_spec->{unwind_prod_dir}.old &");
    }
    unless(eval {mkpath($inst_spec->{unwind_prod_dir},0777);}) {
	upderr_log(0, "mkpath($inst_spec->{unwind_prod_dir}): $!\n");
    }

    unless( eval 'chdir($inst_spec->{unwind_prod_dir})' ) {
	upderr_log(0, "chdir($inst_spec->{unwind_prod_dir}): $!\n");
	return 0;
    }

    if ( -f $dest ) {
	if ($g_flags{opt_O} =~ /clone/ ) {
	   upderr_log(0,"Error: destination product exists");
           return 0;
        } else {
	   upderr_log(0,"Notice: removing previous copy of product tarfile\n");
	   unlink($dest);
        }
    }

    my $ftphome;
    $ftphome = (glob('~ftp'))[0];
    $src = "$ftphome/$inst_spec->{dist_archive_file}";
    if ( ! -f $src ) {
        $src = $inst_spec->{dist_archive_file};
    }
    
    unless (copy($src, $dest) ) {
	
	upderr_log(0, "copy($src, $dest): $!\n");
	return 0;
    }

    if ( defined $updxfr::unpack_command{$suffix} ) {
	if ( updusr::updusr_unwind_all($inst_spec, $g_flags{opt_O}) ) {
	    $cmd = sprintf("$updxfr::unpack_command{$suffix}", $dest);
	} else {
	    $cmd = sprintf("$updxfr::unpack_command{$suffix} $unpack", $dest);
	}

        upderr_log(2, "running: $cmd\n");

	if ( 0 != system( $cmd )) {
	    upderr_log(0, "system($cmd): $!\n");
            # unlink($src) || upderr_log(0, "Notice, unable to remove $src\n");
	    # return 0;
	}
    }

    unless ($g_flags{opt_O} =~ m/clone/ ) {
        unlink($src) || upderr_log(0, "Notice, unable to remove $src\n");
        unlink($inst_spec->{dist_archive_file});
    }


    $inst_spec->{archive_file} = $ddest;

    # updusr::updusr_post_declare($inst_spec, $g_flags{opt_O});
    upderr_get(); upderr_clear();

    print $inst_spec->{database}, "\n";
    if (defined ($inst_spec->{archive_file})) {
        print $inst_spec->{archive_file}, "\n";
    } else {
	print "\n";
    }
    print $inst_spec->{prod_dir}, "\n";
    return 1;
}

sub updcmd_move_table_file {
    my ($inst_spec) = @_;
    my $ok;

    upderr_log(3, "entering updcmd_move_table_file\n" );

    return 0 unless (updcmd_get_move_locations($inst_spec));

    upderr_log(3, "cd-ing to product dir\n" );
    if ( ! -d $inst_spec->{unwind_prod_dir}) {
        upderr_log(3, "making unwind_prod_dir $inst_spec->{unwind_prod_dir}...\n" );
        unless (eval {mkpath($inst_spec->{unwind_prod_dir},0777);}) {
	    upderr_log(0, "mkpath($inst_spec->{unwind_prod_dir}): $!\n");
	    return 0;
	}
    }

    unless (chdir($inst_spec->{unwind_prod_dir})) {
	upderr_log(0, "chdir($inst_spec->{unwind_prod_dir}): $!\n");
	return 0;
    }

    if ( ! -d $inst_spec->{unwind_table_dir}) {
        upderr_log(3, "making table_dir $inst_spec->{unwind_table_dir}...\n" );
        unless (eval {mkpath($inst_spec->{unwind_table_dir},0777);}) {
	    upderr_log(0, "mkpath($inst_spec->{unwind_table_dir}): $!\n");
	    return 0;
	}
    }

    my $ftphome;
    $ftphome = (glob('~ftp'))[0];
    my $src;
    
    $src = "$inst_spec->{dist_table_dir}/$inst_spec->{dist_table_file}";
    if ( ! -f ${src} ) {
	$src = "$ftphome/$src";
    }
    my $dst = "$inst_spec->{unwind_table_dir}/$inst_spec->{table_file}";

    if ( $src ne $dst ) {

	upderr_log(3, "copying table_file $dst...\n" );
	if (copy($src,$dst . ".new")) {
	    link($dst, $dst . ".old");
	    rename($dst . ".new", $dst);

	    # if this is a modify, do a verify with the new table file
	    if ($g_flags{opt_O} =~ m/modify/) {
	       updups_verify(0,$inst_spec,\$ok);
	    } else {
	       $ok = 1;
	    }

	    # if the verify isn't clean, put back the old table file
	    if ( !$ok ) {
		upderr_log(0,"ups verify failed with new table file, putting back old one");
		rename($dst . ".old", $dst);
	    }
	} else {
	    upderr_log(0, "copy($src,$dst");
	    upderr_log(0, "unable to copy table file:$!\n");
	    unlink($src) || upderr_log(0, "Notice, unable to remove $src\n");
	    unlink($dst. ".new");
	    upderr_get(); upderr_clear();
	    return 0;
	}

	unless ($g_flags{opt_O} =~ /clone/ ) {
	  unlink($src) || upderr_log(0, "Notice, unable to remove $src\n");
	}
    }


    upderr_get(); upderr_clear();

    print $inst_spec->{database} , "\n";
    print $inst_spec->{table_file} , "\n";
    print $inst_spec->{table_dir} , "\n";
    return $ok;
}

sub updcmd_move_ups_dir {
    my ($inst_spec) = @_;

    upderr_log(3, "entering updcmd_move_table_file\n" );

    return 0 unless (updcmd_get_move_locations($inst_spec));

    upderr_log(3, "cd-ing to product dir\n" );
    if ( ! -d $inst_spec->{unwind_prod_dir}) {
        upderr_log(3, "making prod_dir $inst_spec->{unwind_prod_dir}...\n" );
        unless (eval {mkpath($inst_spec->{unwind_prod_dir},0777);}) {
	    upderr_log(0, "mkpath($inst_spec->{unwind_prod_dir}): $!\n");
	    return 0;
	}
    }

    unless (chdir($inst_spec->{unwind_prod_dir})) {
	upderr_log(0, "chdir($inst_spec->{unwind_prod_dir}): $!\n");
	return 0;
    }

    my ($ftphome, $src, $cmd);

    if ( -d $inst_spec->{dist_ups_dir} && $g_flags{opt_O} =~ m/clone/ ) {
        $src = upduti_make_tarfile($inst_spec->{dist_ups_dir});
    } else {
        $ftphome = (glob('~ftp'))[0];
        $src = "$ftphome/$inst_spec->{dist_ups_dir}";
        if ( ! -f $src ) {
	    $src = $inst_spec->{dist_ups_dir};
	}
    }
    if ( ! -f $src ) {
	upderr_log(0, "Error: no tarfile of ups directory");
	return 0;
    }

    my $newups = $inst_spec->{unwind_ups_dir} . '.new';

    if ( ! -d $newups) {
        upderr_log(3, "making ups_dir $newups...\n" );
        unless (eval {mkpath($newups,0777);}) {
	    upderr_log(0, "mkpath($newups): $!\n");
	    return 0;
	}
    }

    unless (chdir($newups)) {
	upderr_log(0, "chdir($newups): $!\n");
	return 0;
    }


    $cmd = sprintf("$updxfr::unpack_command{'tar'}", $src);

    if ( 0 != system( $cmd )) {
	upderr_log(0, "system($cmd): $!\n");
        unlink($src) || upderr_log(0, "Notice: unable to unlink $src\n");
	return 0;
    }

    unless (chdir($inst_spec->{unwind_prod_dir})) {
	upderr_log(0, "chdir($inst_spec->{unwind_prod_dir}): $!\n");
	return 0;
    }

    rename ($inst_spec->{unwind_ups_dir},$inst_spec->{unwind_ups_dir}.'.bak');
    rename ($newups, $inst_spec->{unwind_ups_dir});
    system ('rm -rf ' . $inst_spec->{unwind_ups_dir} . '.bak &');

    unlink($src) || upderr_log(0, "Notice: unable to unlink $src\n");
    upderr_get(); upderr_clear();

    print $inst_spec->{database} , "\n";
    print $inst_spec->{ups_dir} , "\n";
    return 1;
}

sub updcmd_get_proprietarylist {
    my($dist_url,$dist_db) = @_;
    local $updcmd::proprietarylist;
    my $prodhash;

    upderr_log(3,"entering, for node $dist_url->{node}\n");

    if (!defined $upduti::proprietarylist) {

	$prodhash = { 
		'product' => 'proprietarylist',
		'flavor' => 'NULL',
		'version' => 'only',
	};

	if (defined $dist_db) {
	    $prodhash->{database} = $dist_db;
	}


	updups_find($dist_url->{str}, $prodhash, '_proprietarylist');

	$updcmd::proprietarylist = $prodhash->{'_proprietarylist'};
    }
    upderr_log(3,"proprietarylist is $updcmd::proprietarylist\n");
    return ":" . $updcmd::proprietarylist . ":";
}

##########################################################################
# Private functions
#

#-------------------------------------------------------------------------
# <return> = <function name>( <input> )
#
# <description>
#
# Input : <input>
# Output: <output>
# Return: <return>
#
#sub <function name>
#{
#}
# If standalone, run tests
if ( $0 =~ /updcmd.pm$/ )
{
my $test_usage = "usage: perl updcmd.pm -h <upd_distribution_node> [-vvv]\n";
my $rhost = '';
my $verbose = 0;

# parse arguments

while( ($_ = shift) ) {
  if ( /^-h/ ) { $rhost = shift; }
  elsif ( /^-v/ ) { $verbose = length( $_ ) -1; }
  else { print $test_usage; exit 1; }
}

# we need this one

if ( ! $rhost ) {
  print $test_usage;
  exit 1;
}

upderr_setverbose($verbose);

my %a_inst = (
           product     => "exmh",
           flavor      => "IRIX",
           version     => 'v1_6_6',
           qualifiers  => '',
           database    => "/blah2",
           foundby     => "current"
           );

my $ret_val;

$g_flags{opt_h} = $rhost;
$g_flags{opt_s} = "1";
$ret_val = updcmd_install(\%a_inst, 0);
print "main: output from updcmd_install = $ret_val\n";
upderr_get(); upderr_clear();

}

1;

__END__

=head1 NAME

updxyz - blah blah

=head1 SYNOPSIS

<use it this way>

=head1 DESCRIPTION

<description>

=head1 BUGS

<bugs>

=cut
