##########################################################################
#
# FILE:
#       upderr.pm
# 
# DESCRIPTION: 
#       Functions to handle the upd error messages.
#
# AUTHORS:
#       Marc Mengel
#       Lars Rasmussen
#       Margaret Votava
#       Don Walsh
#
#       Fermilab Computing Division
#       Batavia, Il 60510, U.S.A.                                              
#
# MODIFICATIONS:
#	28-jul-1998, llc, add $ERRWIDTH and upderr_setwidth;
#	                  modify upderr_log so that it handles
#	                  lines of arbitrary initial width and
#                         breaks them into chunks no larger than
#                         $ERRWIDTH
#       dd-mmm-yyyy, NN, first
#
##########################################################################

package upderr;
use updgbl;


require Exporter;

@ISA = qw(Exporter);

##########################################################################
# Exported Functions
#
##########################################################################

@EXPORT = qw ( upderr_syslog
	      	   upderr_log
	      	   upderr_get
	      	   upderr_setverbose
	      	   upderr_getverbose
		   upderr_setwidth
	      	   upderr_clear
	      	);

##########################################################################
# Global and local variables
#
##########################################################################

my $ERRMSGS           = '';		# store error messages here till printed
my $CURRENT_VERBOSITY = 0;		# Contains the current verbosity level.
my $HAVE_SYSLOG       = 0;

my $ERRWIDTH = 80;			# Contains the desired width of output messages

##########################################################################
# What to use
#
##########################################################################

##########################################################################
# BEGIN
# check if syslog.ph is in the Perl's library path (taken from require).
#
sub BEGIN {

my $fname = 'syslog.ph';

#  if ( $INC{ $fname } ) {
#    $upderr::HAVE_SYSLOG = 1;
#    return;
#  }

  my $fpath = '';  
  LOOP: {
    foreach $pfix ( @INC ) {
      $fpath = "$pfix/$fname";
      if ( -f $fpath ) {
		$upderr::HAVE_SYSLOG = 1;
		last LOOP;
      }
    }
  }

  if ( ! $upderr::HAVE_SYSLOG ) {
    print STDERR "WARNING: could not find syslog.ph, needed by upderr_syslog\n";
  }
}

use strict;
use integer;
use Cwd;

if ( $upderr::HAVE_SYSLOG )
   {
   require Sys::Syslog;
   import Sys::Syslog;
   }

#-------------------------------------------------------------------------
# upderr_syslog(scalar $priority, $message )
#
# Writes messages to the system log file 'syslog'.
#
# Input : scalar, $priority
#		: scalar, $message
# Output: none
# Return: none
#

sub upderr_syslog {
my	($priority,
	 $message) = @_;

      if ($g_flags{opt_s}) {
            return;
      }
      if ( $upderr::HAVE_SYSLOG ) {
		openlog('upderr', 'cons,pid', 'user');	# Open the syslog file
		syslog($priority, $message); 			# write message to the file
		closelog();                             # Close the file
      }
      # we used to always log it to the user anyway, but I'm now
      # adding specific, more breif, upderr_log messages
      # upderr_log(0, $message);
      return;		
}

#-------------------------------------------------------------------------
# upderr_log(scalar verbosity, $message)
#
# Log messages to the user. Verbosity is the level of verbosity
# that needs to be set to see the message. Messages with a
# verbosity of '0' are always printed.
#
# Input : scalar, verbosity
#       : scalar, $message
# Output: none
# Return: none
#

sub upderr_log {
my	($verbosity,
	 $message) = @_;
        my ($i);

        $i = 1;
        while( caller($i) ) {
          $i++;
        }
	my $funcname;
	$funcname = (caller(1))[3];
	$funcname =~ s/.*://;
	$funcname = substr("            ",0,$i) . $funcname;

        chomp($message);

	my ($n);
        if ($CURRENT_VERBOSITY > 0) {
  	    $message = $funcname . " - " . $message . "\n";
	    $n = 0;
	    while ( length($message) > (($n+1)*$ERRWIDTH + ($n-1)) ) {
		$message = substr($message, 0, (($n+1)*$ERRWIDTH + ($n-1))) .
		           "\\\n" .
		           $funcname . 
		           "   " . 
			   substr($message, (($n+1)*$ERRWIDTH + ($n-1)));
		$n++;
	    }
	} else {
            if ( $message ne '') {
  	         $message = $message . "\n";
            }
	}
	
	if ( $verbosity < 0 ) { $verbosity = 0; }
	if ( $verbosity == 0) {
		$ERRMSGS .= "$message";
	} elsif ($verbosity <= $CURRENT_VERBOSITY ) {
		$ERRMSGS .= "$message";
        }
       return;		
}

#-------------------------------------------------------------------------
# upderr_get(strings)
#
# Collect messages reported so far with upderr_log(). Usually
# called by someone about to print the resulting string.
#
# Input : scalar errstr
# Output: none
# Return: none
#
sub upderr_get {
my	($errstr ) = @_;
	
my @err_string = ();
	
	if (ref ($errstr)) {			# return the contents of ERRMSGS
		upderr_log(2, "Returning entire errmsgs\n");
		${$errstr} = $ERRMSGS;
		return;
	}
	
	print STDERR $ERRMSGS;
	return;		
}

#-------------------------------------------------------------------------
# upderr_setverbose(verbosity)
#
# Sets the level of verbosity. Always return the
# current level.
#
# Input : scalar verbosity
# Output: none
# Return: current verbose level
#
sub upderr_setverbose {
my	($verbosity ) = @_;

	$CURRENT_VERBOSITY = $verbosity;
	return $CURRENT_VERBOSITY;
}

sub upderr_getverbose {

	return $CURRENT_VERBOSITY;
}

#-------------------------------------------------------------------------
# upderr_setwidth()
#
# Sets the global ERRWIDTH, width of messages to be logged
#
# Input : scalar width
# Output: none
# Return: current width of screen
sub upderr_setwidth {
my	($errwidth ) = @_;

	$ERRWIDTH = $errwidth;
	return $ERRWIDTH;
}

#-------------------------------------------------------------------------
# upderr_clear()
#
# Sets the global $ERRMSGS ''
# current level.
#
# Input : none
# Output: none
# Return: none
#
sub upderr_clear {

	$ERRMSGS = '';
}


1;

__END__

=head1 NAME

upderr - blah blah

=head1 SYNOPSIS

<use it this way>

=head1 DESCRIPTION

<description>

=head1 BUGS

<bugs>

=cut


