##########################################################################
#
# FILE:
#       updstb.pm
# 
# DESCRIPTION: 
#       This module contain routines that are stubs for the user callouts
#       to make the callouts as simple as possible
#
# AUTHORS:
#       Marc Mengel
#       Lars Rasmussen
#       Margaret Votava
#       Don Walsh
#
#       Fermilab Computing Division
#       Batavia, Il 60510, U.S.A.                                              
#
# MODIFICATIONS:
#       dd-mmm-yyyy, NN, first
#
##########################################################################
package updstb;

use upderr;
use updchk;
use updups;
use updgbl;
use strict;
use integer;
use Exporter();
use vars qw( @ISA @EXPORT );
@ISA = qw( Exporter );

# export this !
@EXPORT = qw (
 	updstb_get_db
 	updstb_get_proddir
	updstb_predeclare
	updstb_postdeclare
	updstb_get_tablefile
	updstb_get_upsdir
	);

BEGIN {
    my $dir;
    my $found = 0;
    push(@INC,'.');
    foreach $dir ( "$ENV{HOME}", "$ENV{UPD_USERCODE_DIR}" ) {
        if ( -f "$dir/updusr.pm" ) {
	    #
	    # note that this hardly ever happens 'cause verbosity isn't set
	    # yet ..
	    #
            upderr_log (2,"found updusr in $dir\n");

            require "$dir/updusr.pm";
            import updusr;
            $found = 1;
            last;
        }
    }
    if ( !$found ) { print STDERR  "Notice: no updusr.pm module found\n";}
    pop(@INC);
    1;
}
##########################################################################
# Exported functions
#


#-------------------------------------------------------------------------
# status = updstb_get_db ( \%inst_spec, \%top_inst_spec)
#
# Gets and stores (into my_instance_spec) the database into which the
# product should be declared.
#
# Input : hash, inst_spec	instance to get db for
#         hash, top_inst_spec	for overlaid products, this is the main
#                               product, otherwise it doesn't matter.
#         scalar, dist_db	distribution database
# Output: none
# Return: scalar, status
#
sub updstb_get_db{
my ($inst_spec, $top_inst_spec) = @_;

upderr_log(1,"updstb_get_db($inst_spec,$top_inst_spec)\n");# debug info

# first make sure instance spec is filled out, and save important stuff
# ---------------------------------------------------------------------
return 0 unless updchk_instance_spec($inst_spec,"unique");
return 0 unless updchk_instance_spec($top_inst_spec,"unique");
$inst_spec->{dist_database} = $inst_spec->{database};		# save dist db
delete $inst_spec->{database};                                  # remove local db
# make user callout
# -----------------
unless (updusr_get_db($inst_spec, $top_inst_spec, $g_flags{opt_O}))
   {
   upderr_log(0,"ERROR: User callout error.\n");
   return 0;
   }

upderr_log(3,"   database = $inst_spec->{database}\n");		# debug info
return 1;
}


#-------------------------------------------------------------------------
# status = updstb_get_proddir( \%real_inst_spec, \%top_inst_spec) 
#
# Gets and stores (into my_instance_spec) the directory into which the
# product files should be unwound and declared (these might be separated).
#
# Input : hash, real_inst_spec	instance to get dir for
#         hash, top_inst_spec	for overlaid products, this is the main
#                               product, otherwise it doesn't matter.
# Output: none
# Return: scalar, status
#
sub updstb_get_proddir{
my ($real_inst_spec, $top_inst_spec) = @_;

upderr_log(1,"updstb_get_proddir($real_inst_spec,$top_inst_spec)\n");

# first make sure instance spec is filled out
# -------------------------------------------
return 0 unless updchk_instance_spec($real_inst_spec,"unique");
return 0 unless updchk_instance_spec($top_inst_spec,"unique");

unless(defined( $real_inst_spec->{dist_prod_dir})) # save dist dir
   {
   $real_inst_spec->{dist_prod_dir} = $real_inst_spec->{prod_dir};
   delete $real_inst_spec->{prod_dir};
   }

unless(defined( $real_inst_spec->{dist_archive_file})) # save dist archive file
   {
   $real_inst_spec->{dist_archive_file} = $real_inst_spec->{archive_file};
   delete $real_inst_spec->{archive_file};
   }

# make user callout
# -----------------
unless (updusr_get_proddir($real_inst_spec, $top_inst_spec,$g_flags{opt_O}))
   {
   upderr_log(0,"ERROR: User callout error.\n");
   return 0;
   }

unless (defined ($real_inst_spec->{unwind_prod_dir}))
   {
   upderr_log(0,"ERROR: unwind_prod_dir undefined for " .
      "$real_inst_spec->{product}\n");
   return 0;
   }
unless (defined ($real_inst_spec->{prod_dir}))
   {
   upderr_log(0,"ERROR: prod_dir undefined for $real_inst_spec->{product}\n");
   return 0;
   }
upderr_log(3,"   dist product dir = $real_inst_spec->{dist_prod_dir}\n");
upderr_log(3,"   product dir = $real_inst_spec->{prod_dir}\n");
upderr_log(3,"   unwind product dir = $real_inst_spec->{unwind_prod_dir}\n");
return 1;
}


#-------------------------------------------------------------------------
# status = updstb_predeclare( $inst_spec )
#
# Release a volume, etc (if using the afs file system)
#
# Input : scalar, dir		directory to release
# Output: none
# Return: scalar, status
#
sub updstb_predeclare {
my ($inst_spec) = @_;
upderr_log(1,"updstb_predeclare($inst_spec)\n");

if ( defined($g_flags{opt_s}) && $g_flags{opt_s} ) {
    upderr_log(0, "I would run the predeclare user callout now.");
    return 1;
}
# make user callout
# -----------------
unless (updusr_predeclare($inst_spec,$g_flags{opt_O}))
   {
   upderr_log(0,"ERROR: User callout error.\n");
   return 0;
   }

return 1;
}

#-------------------------------------------------------------------------
# status = updstb_postdeclare( $inst_spec )
#
# Release a volume, etc (if using the afs file system)
#
# Input : scalar, dir		directory to release
# Output: none
# Return: scalar, status
#
sub updstb_postdeclare {
my ($inst_spec) = @_;
upderr_log(1,"updstb_postdeclare($inst_spec)\n");

if ( defined($g_flags{opt_s}) && $g_flags{opt_s} ) {
    upderr_log(0, "I would run the postdeclare user callout now.");
    return 1;
}

# make user callout
# -----------------
unless (updusr_postdeclare($inst_spec,$g_flags{opt_O}))
   {
   upderr_log(0,"ERROR: User callout error.\n");
   return 0;
   }

return 1;
}


#-------------------------------------------------------------------------
# status = updstb_get_tablefile ( \%inst_spec)
#
# get table file location of product instance
#
# Input : hash, inst_spec	
# Output: none
# Return: scalar, status
#
sub updstb_get_tablefile{
my ($inst_spec) = @_;

upderr_log(1,"updstb_get_tablefile($inst_spec)\n");

# first make sure instance spec is filled out
# -------------------------------------------
return 0 unless updchk_instance_spec($inst_spec,"unique");
$inst_spec->{dist_table_dir} = $inst_spec->{table_dir};
$inst_spec->{dist_table_file} = $inst_spec->{table_file};
delete $inst_spec->{table_file};
delete $inst_spec->{table_dir};

# make user callout
# -----------------
unless (updusr_get_tablefile($inst_spec,$g_flags{opt_O}))
   {
   upderr_log(0,"ERROR: User callout error.\n");
   return 0;
   }

unless (defined ($inst_spec->{table_file}))
   {
   upderr_log(0,"ERROR: table_file undefined for $inst_spec->{product}\n");
   return 0;
   }
unless (defined ($inst_spec->{unwind_table_dir}))
   {
   upderr_log(0,"ERROR: unwind_table_dir undefined for " .
      "$inst_spec->{product}\n");
   return 0;
   }
upderr_log(3,"   table file = " .
      "$inst_spec->{unwind_table_dir}/$inst_spec->{table_file}\n");
return 1;
}


#-------------------------------------------------------------------------
# status = updstb_get_upsdir( \%inst_spec)
#
# get ups subdirectory of product instance
#
# Input : hash, inst_spec	
# Output: none
# Return: scalar, status
#
sub updstb_get_upsdir{
my ($inst_spec) = @_;
upderr_log(1,"updstb_get_upsdir($inst_spec)\n");

# first make sure instance spec is filled out
# -------------------------------------------
return 0 unless updchk_instance_spec($inst_spec,"unique");
$inst_spec->{dist_ups_dir} = $inst_spec->{ups_dir};
delete $inst_spec->{ups_dir};

# make user callout
# -----------------
unless (updusr_get_upsdir($inst_spec,$g_flags{opt_O}))
   {
   upderr_log(0,"ERROR: User callout error.\n");
   return 0;
   }

unless (defined ($inst_spec->{unwind_ups_dir}))
   {
   upderr_log(0,"ERROR: unwind_ups_dir undefined for " .
      "$inst_spec->{product}\n");
   return 0;
   }
upderr_log(3,"   ups dir = $inst_spec->{unwind_ups_dir}\n");
return 1;
}


##########################################################################
# Private functions
#
# If standalone, run tests
if ( $0 =~ /updstb.pm$/ )
{
my %a_inst = (
           product     => "exmh",
           flavor      => "IRIX",
           version     => 'v1_3',
           qualifiers  => '',
           database    => "blah",
           foundby     => "current"
           );
my %b_inst = (
           product     => "tcl",
           flavor      => "IRIX",
           version     => 'v2_2',
           qualifiers  => '',
           foundby     => "current",
           database    => "blah",
           );
my %c_inst = (
           product     => "tk",
           flavor      => "IRIX",
           version     => 'v3_0',
           qualifiers  => '',
           foundby     => "current",
           database    => "blah",
           );
my %d_inst = (
           product     => "exmh",
           flavor      => "IRIX",
           version     => 'v1_2',
           qualifiers  => '',
           foundby     => "current",
           database    => "blah",
           );
my %e_inst = (
           product     => "tcl",
           flavor      => "IRIX",
           version     => 'v2_2',
           qualifiers  => '',
           foundby     => "current",
           database    => "blah",
           );
my %f_inst = (
           product     => "tk",
           flavor      => "IRIX",
           version     => 'v3_0',
           qualifiers  => '',
           foundby     => "current",
           database    => "blah",
           );
my %g_inst = (
           product     => "ghostview",
           flavor      => "IRIX",
           version     => 'v3_0',
           qualifiers  => '',
           foundby     => "current",
           database    => "blah",
           );

my $ret_val;
my $proddir = ".";
my $dist_node = "upd";
my $dist_db = "/usr/products/declared";

upderr_setverbose(10);
$ret_val = updstb_get_db( \%a_inst, \%b_inst, $dist_node, $dist_db);
print "main: output from updstb_get_db = $ret_val\n";
upderr_get(); upderr_clear();

$ret_val = updstb_get_proddir( \%a_inst, \%b_inst,
   $dist_node, $dist_db);
print "main: output from updstb_get_proddir = $ret_val\n";
upderr_get(); upderr_clear();

$ret_val = updstb_predeclare( \%a_inst );
print "main: output from updstb_predeclare = $ret_val\n";
upderr_get(); upderr_clear();

$ret_val = updstb_postdeclare( \%a_inst );
print "main: output from updstb_postdeclare = $ret_val\n";
upderr_get(); upderr_clear();

$ret_val = updstb_get_tablefile( \%a_inst );
print "main: output from updstb_get_tablefile= $ret_val\n";
upderr_get(); upderr_clear();

$ret_val = updstb_get_upsdir( \%a_inst );
print "main: output from updstb_get_upsdir = $ret_val\n";
upderr_get(); upderr_clear();
}

1;

__END__

=head1 NAME

updchk - This module contain routines that are used to check consistency 
prerequisites

=head1 SYNOPSIS

<use it this way>

=head1 DESCRIPTION
This module contain routines that are used to check consistency 
prerequisites

<description>

=head1 BUGS

<bugs>

=cut
