##########################################################################
#
# FILE:
#       updupr.pm
# 
# DESCRIPTION: 
#       This module contains routines to encapsulate ups on the 
#       distribution host.
#
# AUTHORS:
#       Marc Mengel
#       Lars Rasmussen
#       Margaret Votava
#       Don Walsh
#
#       Fermilab Computing Division
#       Batavia, Il 60510, U.S.A.                                              
#
# MODIFICATIONS:
#       19-dec-1997, lr, first
#       19-feb-1998, lr, now using LWP
#
##########################################################################
package updupr;

use strict;
use HTTP::Status;
use HTTP::Response;
use LWP::UserAgent;
use Net::Domain qw(hostfqdn);
use Socket;
use Exporter();

use upderr;
use upduti;

# to make strict happy
use vars qw(@ISA @EXPORT);
@ISA = qw(Exporter);

# export this !
@EXPORT = qw (
	      updupr_oxc

	      );

##########################################################################
# Exported variables
#

##########################################################################
# Global and local variables
#
my $g_prog = "erupd/v0.0";
my $g_def_cgi_path = '/cgi-bin';
my $g_def_cgi_ups = 'ups.cgi';

##########################################################################
# Exported functions
#

#-------------------------------------------------------------------------
# $ret_code = updupr_oxc( $url, $ups_cmd, \@stdout_arr, \@stderr_arr )
#
# Will open, execute an ups command and close connection.
#
# Note: it return 0 on success, the ups return code on a ups error, 
#       else -1.
#
# Input : scalar, url: http://<node>:<port>/<path>
#         scalar, ups_command (without 'ups')
# Output: array ref, data, if not passed, data will go to stdout
#         array ref, errors, if not passed, errors will go to stderr
# Return: OS return code from ups
#
sub updupr_oxc {

  my ($url_str, $ups_cmd, $stdout_arr, $stderr_arr) = @_;
  my $ret_code = -1;

  # parse data, if nothing to do, quit

  if ( ! $ups_cmd ) {
    upderr_log( 0, "ERROR: Empty ups command\n" );
    return $ret_code;
  }
  $ups_cmd =~ s/ /&/g;

  # parse URL

  my $url = upduti_parse_url( $url_str );
  return $ret_code unless ( $url );
  $url_str = $url->{str};

  # set default path

  if ( ! $url->{path} ) {
    $url_str .= "$g_def_cgi_path/$g_def_cgi_ups";
  }
  elsif ( $url->{path} =~ /\/$/ ) {
    $url_str .= "$g_def_cgi_ups";
  }

  # create a user agent

  my $ua = new LWP::UserAgent;
  $ua->agent( "$g_prog" );
  $ua->timeout(360);		# turn the timeout up to an hour

  # request a URL

  upderr_log( 1, "$url_str?$ups_cmd\n");

  my $request = new HTTP::Request( "GET", "$url_str?$ups_cmd" );
  $request->authorization_basic($ENV{'USER'}, $ENV{USER} . '@' . &hostfqdn());
  my $response = $ua->request( $request );

  # check response and content type

  if ( $response->code != RC_OK ) {
    upderr_log( 0, "while trying to contact $url->{node}:\n" );
    upderr_log( 0, 'Received HTTP error (' . $response->code . '), ' . $response->message . "\n" );
    return $ret_code;
  }
  elsif ( $response->content_type !~ m@text/plain@ ) {
    upderr_log( 0, 'ERROR: wrong content type, we don\'t do ' . $response->content_type . "\n" );
    exit $ret_code;
  }

  # get the data and errors/warnings

  my @content = split "\n", $response->content;

  # first, what the server did

  my $ups_cmd_str = shift @content;
  upderr_log( 3, "$ups_cmd_str\n");

  # next, data
  # stdout and stderr are separated by a 'UPSR_RETURN_CODE = code'

  my $err_on = '';
  foreach ( @content ) {
    chomp $_;
    upderr_log( 4, "$_\n" );
    if ( /^UPSR_RETURN_CODE/ ) {
      ($err_on, $ret_code) = split / *= */, $_, 2;
      $ret_code = int( $ret_code );
      upderr_log(2,"Got return code of $ret_code\n");
      next;
    }
    if ( $err_on ) {
      if ( defined $stderr_arr ) { push @$stderr_arr, $_; }
      else { print STDERR "$_\n"; } 
    }
    else { 
      if ( defined $stdout_arr ) { push @$stdout_arr, $_; }
      else { print STDOUT "$_\n"; } 
    }
  }

  return $ret_code;
}

##########################################################################
# Private functions
#

##########################################################################
# BEGIN/END
#
# Like to put'm here so everything else are defined.
#

sub BEGIN
{
}

1;

__END__

=head1 NAME

 updupr - This module contains routines to encapsulate ups on the 
distribution host.


=head1 SYNOPSIS

=head1 DESCRIPTION

 This module contains routines to encapsulate ups on the 
distribution host.

=head1 BUGS

 it's not done at all

=cut
