##########################################################################
#
# FILE:
#       updups.pm
# 
# DESCRIPTION: 
#       This module contains routines to encapsulate ups on the local and
#       remote host.
#       (for more info, see pod at the end).
#
# AUTHORS:
#       Marc Mengel
#       Lars Rasmussen
#       Margaret Votava
#       Don Walsh
#
#       Fermilab Computing Division
#       Batavia, Il 60510, U.S.A.                                              
#
# MODIFICATIONS:
#       12-dec-1997, lr, first
#	06-feb-1998, psr, second
#	06-feb-1998, lr, third
#
##########################################################################
package updups;

use strict;
use Exporter();

use updchk;
use upderr;
use updupr;
use updgbl;
use upduti;

# to make strict happy
use vars qw( @ISA @EXPORT 
	   );
@ISA = qw(Exporter);

# export this !
@EXPORT = qw (
	      updups_check_table
	      updups_touch
	      updups_declare
	      updups_undeclare
	      updups_depend
	      updups_verify
   	      updups_list_depend_K 
	      updups_exist
	      updups_find
	      updups_flavor
	      updups_get

	      ups_oxc
	      ispec2str
	      );

##########################################################################
# Exported variables
#

##########################################################################
# Global and local variables
#

# Dedault components for updups_find (local and remote).

my @g_def_comp_loc = ('database');
my @g_def_comp_rem = ('database', 'archive_file', '_upd_overlay');

# Map between instance spec components and ups options used by upd.

my %g_comp2opt = (
		  'flavor'            => 'f:',
		  'qualifiers'        => 'q:',
		  'chain'             => 'g:',
		  'authorized_nodes'  => 'A:',
		  'origin'            => 'D:',
		  'statistics'        => 'L',
		  'table_file'        => 'm:',
		  'table_dir'         => 'M:',
		  'prod_dir'          => 'r:',
		  'archive_file'      => 'T:',
		  'ups_dir'           => 'U:',
		  'database'          => 'z:',
#		  'description'	      => 'p:',
		  'ups_flags'	      => 'O:',
		  'opt_a'	      => 'a',
		  );


# ups options to be considered for different ups command line constructs.

my $g_sopt_unique  = 'fqz';
my $g_sopt_find    = 'fqgz';
my $g_sopt_declare = 'fpOqgADLmMrTUz';

##########################################################################
# Exported functions
#
#-------------------------------------------------------------------------
# $status = updups_check_table( $url, \%instance_spec, \%x_args )
#
# Does a 'ups setup' of instance_spec overriden with x_args ,
#   with some clean up, in order to make sure the table file
#   is compatable with these arguments.
#
# Input : scalar, url, can be empty or local
#         hash, instance_spec
# Output: none
# Return: status, 0 if any errors
#
sub updups_check_table {
  
  my ($url, $inst_spec) = @_;
  my $inst_str = '';
  my $var = '';
  my $opt = '';
  my $status;
  $opt .= 's' if ( $g_flags{opt_s} );


  upderr_log( 3 , "input: ($url, $inst_spec)\n");

  # check passed instance spec
  
  return 0 unless updchk_instance_spec( $inst_spec, "unique" );

  # copy so as not to mess with real spec

  $inst_spec = { %{$inst_spec} };

  # delete items that confuse the setup 
  delete $inst_spec->{version};
  delete $inst_spec->{chain};

  # build ups command string

  $inst_str = ispec2str( $inst_spec, "mMU" );

  ## do ups check_table, when available
  my ($error, $data_arr, $err_arr) = do_ups( $url, "setup -j $inst_str", $opt );
 
  if (defined($err_arr) && $#{$err_arr} >= 0) {
      upderr_log(3,"error output is @{$err_arr}\n");
      $status = 0;
  } else {
      $status = 1;
  }
  if ( $data_arr->[0] ne '/dev/null' && -f $data_arr->[0] ) {
      unlink($data_arr->[0]);
  }
  return $status;
}

#-------------------------------------------------------------------------
# $status = updups_touch( $url, \%instance_spec, \%x_args )
#
# Does a 'ups touch' of instance_spec overridden with x_args.
#
# Note: chain is special, it will only be used in the ups touch 
#       command if passed in %x_args.
#
# Input : scalar, url, can be empty or local
#         hash, instance_spec
# Output: none
# Return: status, 0 if any errors
#
sub updups_touch {
  
  my ($url, $inst_spec) = @_;
  my $inst_str = '';
  my $var = '';
  my $opt = '';
  $opt .= 's' if ( $g_flags{opt_s} );


  upderr_log( 3 , "input: ($url, $inst_spec)\n");

  # check passed instance spec
  
  return 0 unless updchk_instance_spec( $inst_spec, "unique" );

  # build ups command string

  $inst_str = ispec2str( $inst_spec, $g_sopt_declare );

  ## do ups touch, when available
  my ($error, $data_arr) = do_ups( $url, "touch $inst_str", $opt );
 
  upderr_syslog( 'notice' , 
		($url ? "on host $url: " : "") .
		($error ? "failed" : "successful" ) . 
		" ups touch $inst_str\n" ) unless ($g_flags{opt_s});
 
  upderr_log(0, 
		($error ? "error:" : "informational:").
		" ups touch on distribution node " .
		($error ? "failed" : "succeeded")) unless ($g_flags{opt_s});

  return 0 if ( $error );
  return 1;
}

#-------------------------------------------------------------------------
# $status = updups_declare( $url, \%instance_spec, \%x_args )
#
# Does a 'ups declare' of instance_spec overridden with x_args.
#
# Note: chain is special, it will only be used in the ups declare 
#       command if passed in %x_args.
#
# Input : scalar, url, can be empty or local
#         hash, instance_spec
#         hash, x_args
# Output: none
# Return: status, 0 if any errors
#
sub updups_declare {
  
  my ($url, $inst_spec, $x_args, $x2_args) = @_;
  my $inst_str = '';
  my $var = '';
  my $opt = '';
  my $line; 
  $opt .= 's' if ( $g_flags{opt_s} );

  if (!defined($x2_args)) {
      $x2_args = '';
  }

  upderr_log( 3 , "input: ($url, $inst_spec, $x_args)\n");

  # is too!!  how else is addproduct going to work...
  # if ( is_remote( $url ) ) {
  #   upderr_log( 0, "ERROR: not allowed on remote host\n" );
  #   return 0;
  # }

  # work on a copy of inst_spec, so we don't confuse upper layers.
  $inst_spec = {%$inst_spec};
    
  # overwrite with x_args

  foreach ( keys %{$x_args} ) {
    $inst_spec->{$_} = $x_args->{$_} if ($x_args->{$_} ne '');
  }
    
  # check passed instance spec
  
  return 0 unless updchk_instance_spec( $inst_spec, "updups_declare" );

  # build ups command string

  $inst_str = ispec2str( $inst_spec, $g_sopt_declare );

  # well, we will take chain if passed in x_args
  # we are now adding it a second time ?
  #if ( defined ($var = $x_args->{chain}) ) {
  #  $var = "\"\"" unless ( $var );
  #  $inst_str .= " -g $var";
  #}

  # do ups declare

  my ($error, $data_arr) = do_ups( $url, "declare $x2_args $inst_str", $opt );

  upderr_syslog( 'notice' , 
                ($url ? "on host $url: " : "") .
		($error ? "failed" :  "successful" ) . 
		" ups declare $x2_args $inst_str\n" ) unless $g_flags{opt_s};
  upderr_log(0,	
		($error ? "error:" : "informational:").
		" ups declare " .
		($error ? "failed" :  "succeeded" )) unless $g_flags{opt_s};

  foreach $line (@$data_arr) {
	upderr_log(0,$line);
  }

  return 0 if ( $error );
  return 1;
}

#-------------------------------------------------------------------------
# $status = updups_undeclare( $url, \%instance_spec, \%x_args )
#
# Does a 'ups undeclare' of instance_spec overridden with x_args.
#
# Note: chain is special, it will only be used in the ups undeclare 
#       command if passed in %x_args.
#
# Input : scalar, url, can be empty or local
#         hash, instance_spec
#         hash, x_args
# Output: none
# Return: status, 0 if any errors
#
sub updups_undeclare {
  
  my ($url, $inst_spec, $x_args, $x2_args) = @_;
  my $inst_str = '';
  my $var = '';
  my $opt = '';
  $opt .= 's' if ( $g_flags{opt_s} );

  if (!defined($x2_args)) {
      $x2_args = '';
  }

  upderr_log( 3 , "input: ($url, $inst_spec, $x_args)\n");

  # overwrite with x_args

  foreach ( keys %{$x_args} ) {
    $inst_spec->{$_} = $x_args->{$_} if ($x_args->{$_} ne '');
  }
    
  # check passed instance spec
  
  return 0 unless updchk_instance_spec( $inst_spec, "unique" ) ||
		updchk_instance_spec( $inst_spec, "unchain" );
  # build ups command string

  $inst_str = ispec2str( $inst_spec, "qgfz" );

  # well, we will take chain if passed in x_args

  #if ( defined ($var = $x_args->{chain}) ) {
  #  $var = "\"\"" unless ( $var );
  #  $inst_str .= " -g $var";
  #}

  # do ups undeclare

  my ($error, $data_arr) = do_ups( $url, "undeclare $x2_args $inst_str", $opt );

  upderr_syslog( 'notice' , 
                ($url ? "on host $url: " : "") .
		($error ? "failed" :  "successful" ) . 
		" ups undeclare $x2_args $inst_str\n" ) unless $g_flags{opt_s};

  upderr_log(0,	
		($error ? "error:" : "informational:").
		" ups undeclare " .
		($error ? "failed" :  "succeeded" )) unless $g_flags{opt_s};

  return 0 if ( $error );
  return 1;
}


#-------------------------------------------------------------------------
# $status = updups_depend( $url, \%instance_spec, \@dependency_list, $ignore_err )
#
# Does a 'ups depend' to determine what the dependencies the product in 
# the instance_spec are.
#
# Note: Who is doing the -G swicth ???
#
# Input : scalar, host:port, can be empty or local
#         hash, instance_spec
# Output: list, dependency_list
# Return: status, 0 if any errors
#
sub updups_depend {

  my ($url, $inst_spec, $dep_list, $ign_err) = @_;
  @$dep_list = ();
  my $opt = '';
  my $opt2;
  $opt .= 'i' if ( $ign_err );

  upderr_log( 3 , "input: ($url, $inst_spec, $dep_list)\n");

  # get instance string from instance spec

  my $inst_str = ispec2str( $inst_spec, $g_sopt_unique, 1 );
  return 0 if ( ! $inst_str );

  # do ups depend
  if (defined $g_flags{opt_R} && $g_flags{opt_R}) {
	$opt2 = '-R';
  } else {
	$opt2 = '';
  }

  my ($error, $data_arr) = do_ups( $url, "depend $opt2 $inst_str", $opt );

  # fill list of instance spec

  my $pushflag = 1;
  foreach ( @{$data_arr} ) {

    if (s/\A[\s|]*\|__//o ) {
        $pushflag = 0;
    }
    
    if ( m/\w/ ) {

      my $dep_spec = str2ispec( $_ );

      if ( defined $dep_spec->{chain} ) { 
	$dep_spec->{foundby} = $dep_spec->{chain}; 
      }
      else { 
	$dep_spec->{foundby} = ""; 
      }

      if ($pushflag) {
         push @$dep_list, $dep_spec;
      } else {
         unshift @$dep_list, $dep_spec;
      }

    }
  }

  upderr_log( 3 , "output: $dep_list\n");

  return 0 if ($error);
  return 1;
}
#-------------------------------------------------------------------------
# $status = updups_list_depend_K( $url, $cmd, \@instance_spec_list, \@list_list, @keywords) 
#
# Does a 'ups list' to determine what the instances of the product in 
# the instance_spec are.
#
# Input : scalar, 			host:port, can be empty or local
#	  scalar,			cmd - can be "list" or "depend"
#         hash, 			instance_spec
# Output: list of instance_specs, 	list_list
# Return: status, 0 if any errors
#
sub updups_list_depend_K {

  my ($url, $cmd, $inst_spec_list, $list_list, @keywords) = @_;
  my ($new_spec,@fields,$keyword,$inst_spec);
  my (@optionsets);

  @$list_list = ();
  my $opt = '';

  upderr_log( 3 , "input: ($url, $inst_spec, $list_list)\n");

  # get instance string from instance spec

  @optionsets=();
  foreach $inst_spec (@$inst_spec_list) {
      my $inst_str = ispec2str( $inst_spec, $g_sopt_unique . 'a' , 0 );
      push(@optionsets, "$inst_str -K " . join(":",@keywords) );
  }

  # do ups list/depend

  my ($error, $data_arr) = do_ups( $url, 
				"$cmd " . join(',', @optionsets) );

  # fill list of instance specs

  foreach ( @{$data_arr} ) {
      upderr_log(4,"line: $_\n");

      $new_spec = {};
      @fields = m/\"((?:[^\"\\]|\\.)*)\"/g;

      foreach $keyword (@keywords) {
 	  $new_spec->{$keyword} = shift(@fields);
      }
      #llc: especially for upc, I need to know the host db
      $new_spec->{"dist_node"} = $url;
      unshift @$list_list, $new_spec;
  }

  upderr_log(3,"Instance_spec list:\n");
  foreach ( @{$list_list} ) {
      upduti_print_instance_spec(3,$_);
  }

  return 0 if ($error);
  return 1;
}

#-------------------------------------------------------------------------
# $status = updups_verify( $url, \%instance_spec, \$verify )
#
# Does a 'ups verify' to determine if a given product instance verifies.
#
# Input : scalar, host:port, can be empty or local
#         hash, instance_spec
# Output: scalar, value
# Return: status, 0 if any errors
#
sub updups_verify {

  my ($url, $inst_spec, $ret_val) = @_;
  $$ret_val = 0;

  upderr_log( 3 , "input: ($inst_spec, $ret_val)\n");

  # get instance string from instance hash

  my $inst_str = ispec2str( $inst_spec, $g_sopt_unique, 1 );
  return 0 if ( ! $inst_str );

  my ($error, $data_arr, $err_arr) = do_ups( $url, "verify -a $inst_str" );
  # kluge alert -- we think less than 2 lines of output is ok
  # more is bad ...
  $$ret_val = scalar(@$err_arr) < 2;

  upderr_log( 4, "$$ret_val data_arr: @$data_arr err_arr @$err_arr");
  return !$error;
}

#-------------------------------------------------------------------------
# $status = updups_exist( $url, \%instance_spec, \$exist )
#
# Does a 'ups list' to determine if a given product instance exist.
#
# Note: if the product exist but disagree in terms of chains to the 
#       instance, it will still return true (calling 'ispec2str'
#       should handle that).
# 
# Input : scalar, host:port, can be empty or local
#         hash, instance_spec
# Output: scalar, value
# Return: status, 0 if any errors
#
sub updups_exist {

  my ($url, $inst_spec, $ret_val) = @_;
  $$ret_val = 0;

  upderr_log( 3 , "input: ($inst_spec, $ret_val)\n");

  # get instance string from instance hash

  if ( $inst_spec->{database} eq "" ) {
      delete $inst_spec->{database}
  }

  my $inst_str = ispec2str( $inst_spec, $g_sopt_unique, 1 );
  return 0 if ( ! $inst_str );

  # execute ups list (probaly 'ups exist -j' should be used here,
  # but currently it returns failer/success in the 'OS return code',
  # which right now doesn't fit into do_ups.

  my ($error, $data_arr) = do_ups( $url, "list -K product $inst_str" );
  #return 0 if ( $error );

  # did we find to much ?

  #
  # If there are multiple instances, it *definitely* exists!!!
  #
  #if ( scalar ( @$data_arr ) > 1 ) {
  #  upderr_log( 0, "ERROR: multiple instances found\n" );
  #  return 0;
  #}

  $$ret_val = 1 if $data_arr->[0];
    
  upderr_log( 3, "output: $$ret_val\n" );

  return 1;
}

#-------------------------------------------------------------------------
# $status = updups_find( $url, \%instance_spec, component-1, component-2 ... )
#
# Does a 'ups list' to obtain the value associated with 
# the components in the product mentioned in instance_spec.
# If no components are passed a default set is used.
#
# Note: It will not overwrite components aready present in instance_spec.
#       Except if a list of databases are passed, in that case it will 
#       always request the database from 'ups list' and overwrite the 
#       instance spec with the database where the instance actually was 
#       found.
#
# Input : scalar, host:port, can be empty or local
#         hash, instance_spec
#         scalars, component-1, ..., component-2
# Output: found components are put into the instance_spec
# Return: status, 0 if any errors, no instance found or to many instances 
#         found.
#
sub updups_find {
  
  my $url = shift;
  my $inst_spec = shift;
  my @comps_in = @_;
  my @comps = ();

  upderr_log( 3 , "input: ($url, $inst_spec, @comps_in)\n");

  # get instance string from instance hash

  my $inst_str = ispec2str( $inst_spec, $g_sopt_find );
  return 0 if ( ! $inst_str );

  # take care of default components

  if ( ! @comps_in ) {
    if ( is_remote( $url ) ) {
      @comps_in = @g_def_comp_rem;
    }
    else {
      @comps_in = @g_def_comp_loc;
    }
  }

  # create the ups -K option and a list of components actually 
  # used for ups list

  my $opt_K = '';
  foreach ( @comps_in ) {
    unless ( defined $inst_spec->{$_} ) {
      $opt_K .= "$_:";
      push @comps, $_;
    }
  }

  # it's special if passed database is a list

  if ( (defined $inst_spec->{database}) && 
       ($inst_spec->{database} =~ /[ :]/) ) {
    
      $opt_K .= 'database:';
      push @comps, 'database';
  } 

  chop ($opt_K);

  return 1 if ( ! $opt_K );

  # do the ups command
  
  my ($error, $data_arr) = do_ups( $url, "list -K $opt_K $inst_str" );
  return 0 if ( $error );

  # did we find to much ?

# This is *not* an error -- ups list will often list multiple
#     ones even when it is not ambiguous to setup.
#
#  if ( scalar ( @$data_arr ) > 1 ) {
#    upderr_log( 0, "ERROR: multiple instances found\n" );
#    return 0;
#  }
    
  my $val = $data_arr->[0];
  chomp( $val ) if ( $val );

  # return false if no instance found

  return 0 if ( ! $val );

  # add values to the instance spec

  my @vals = ( $val =~ m/\s*((?:[^\s"]|"(?:[^"\\]|\\.)*")+)/go );
  
  # consistensy check: number of components should equal number of values

  if ( $#comps != $#vals ) {
    upderr_log( 0, "ERROR: number of returned values not equal number of key words\n" );
    return 0;
  }    
			      
  my $n = 0;
  foreach ( @comps ) {
    $inst_spec->{$_} = $vals[$n++];
    $inst_spec->{$_} =~ s/\A"?(.*?)"?\Z/$1/o;
  }
      
  return 1;
}

#-------------------------------------------------------------------------
# $status = updups_flavor( $url, $opt, \$value )
#
# Does a 'ups flavor' to obtain the machine operating system.
# (usefull options to ups flavor is -0, -1, -2, -3, where
# -0 gives most information)
#
# Input : scalar, host:port, can be empty or local
#         scalar, options to be passed to ups flavor
# Output: scalar, output from ups flavor
# Return: status, 0 if any errors
#
sub updups_flavor {

  my ($url, $opt, $ret_val) = @_;
  $$ret_val = '';

  upderr_log( 3 , "input: ($url, $opt, $ret_val)\n");

  # build command

  my $cmd = "flavor";
  $cmd = "$cmd $opt" if ( $opt );

  # execute ups flavor

  my ($error, $data_arr) = do_ups( $url, $cmd );
  return 0 if ( $error || !$data_arr->[0] );

  $$ret_val = $data_arr->[0];  

  return 1;
}  

#-------------------------------------------------------------------------
# $status = updups_get( $url, \%instance_spec, \@file_list )
#
# Does a 'ups get' to obtain the location of files external to the
# product root.
#
# Input : scalar, host:port, can be empty or local
#         hash, instance spec
# Output: ref to array, list of external files.
# Return: status, 0 if any errors
#
sub updups_get {

  my ($url, $inst_spec, $file_list) = @_;
  @$file_list = ();

  upderr_log( 3 , "input: ( $url, $inst_spec, $file_list)\n");

  # get instance string from instance hash

  my $inst_str = ispec2str( $inst_spec, $g_sopt_unique, 1 );
  return 0 if ( ! $inst_str );

  # execute ups get

  my ($error, $data_arr) = do_ups( $url, "get $inst_str" );
  return 0 if ( $error );

  $file_list = $data_arr;

  return 1;
}

##########################################################################
# Private functions
#

#-------------------------------------------------------------------------
# ($error_code, \@data_arr, \@err_arr) = do_ups( $url, $ups_cmd_line, $opt )
#
# Will determine if it's a local or remote ups command and execute it.
# data is returned as a reference to an array of strings
#
# opt: s - just show what command would do.
#    : i - ignore errors
#
sub do_ups {

  my ($url, $cmd_line, $opt) = @_;
  my @stdout_arr = ();
  my @stderr_arr = ();
  my $ret_code = 0;

  $opt = '' unless ( $opt );

  # here or there ?
  
  $url = 'local' if ( ! is_remote( $url ) );

  # check if 'OS flavor' is passed on command line (-H option)
  # in both local & remote accesses!

  $cmd_line .= " -H $g_flags{opt_H}" if $g_flags{opt_H};

  # execute command

  if ( $url =~ /local/ ) {

    # LOCAL ups
    # =========

    # check if 'only list' option is passed

    if ( $opt =~ /s/ ) {
      upderr_log( 0, "I would do 'ups($url) $cmd_line'\n" );
      return ($ret_code, \@stdout_arr, \@stderr_arr);
    }

    upderr_log( 1, "'ups($url) $cmd_line'\n" );

    $ret_code = ups_oxc( $cmd_line, \@stdout_arr, \@stderr_arr );

  }
  else {

    # REMOTE ups
    # ==========


    # check if 'only list' option is passed

    if ( $opt =~ /s/ ) {
      upderr_log( 0, "I would do 'ups($url) $cmd_line'\n" );
      return ($ret_code, \@stdout_arr);
    }

    upderr_log( 1, "'ups($url) $cmd_line'\n" );

    $ret_code = updupr_oxc( $url, $cmd_line, \@stdout_arr, \@stderr_arr );
  }

  # ignore errors ?
  
  $ret_code = 0 if ( $opt =~ /i/ );

  # log error/warning messages

  # this message looks weird in isolation, but it fits in with what
  # updupr_oxc() prints out...

  upderr_log( 0, "failed to run 'ups $cmd_line' error output is:\n" )
      if ( $ret_code );

  unless ( ($opt =~ /i/) ) {
    foreach ( @stderr_arr ) {
      upderr_log( 0, "$_\n");
    }
  }

  return ($ret_code, \@stdout_arr, \@stderr_arr);
}

#-------------------------------------------------------------------------
# bool = is_remote( $url )
#
# Will just check is $url is local or remote
#
sub is_remote {

  my $url = shift;
  return 0 if ( ! $url || $url =~ /^local/i );
  return 1;
}

#-------------------------------------------------------------------------
# $instance_string = ispec2str( \%instance_spec, $valid_opt, $check_unique )
#
# Will build a string to be parsed by ups.
#
# $valid_opt: list of ups options to be considered in creating the ups 
#             command string.
# $check_unique: if set, passed instance spec will be checked by 
#                updchk_instance_spec( , 'unique' )
# 
sub ispec2str {

  my ($inst_spec, $vopt, $unique) = @_;
  my $inst_str = '';

  # check instance spec

  if ( $unique && ! updchk_instance_spec( $inst_spec, 'unique' ) ) {
    upderr_log( 0, "ERROR: Instance is not unique\n" );
    return 0;
  }
  else {
    # upc needs to be able to do this...
    # if ( ! $inst_spec->{product} ) {
    #   upderr_log( 0, "ERROR: Instance has no product name\n" );
    #   return 0;
    # }      
  }

  $inst_str = "$inst_spec->{product}" if defined $inst_spec->{product};
  $inst_str .= " $inst_spec->{version}" if defined $inst_spec->{version};

  # build an ups command string
  
  foreach my $key ( keys %{$inst_spec} ) {

    next unless defined $g_comp2opt{$key};

    my ($opt, $need_var) = split //, $g_comp2opt{$key};

    # should we use that option ?

    if ( $vopt =~ /$opt/) {

      # an empty variable should be ""

      my $var = $inst_spec->{$key};
      $var = "\"\"" unless ( $var );

      # if variable contains white spaces, add quotes

      if ( $var =~ /\s/ ) {
	my $tmp = "\"";
	$tmp .= $var . "\"";
	$var = $tmp;
      }

      # add ups option to string

      $inst_str .= " -$opt";
      $inst_str .= " $var" if ( $need_var eq ':');
    }
  }

  return $inst_str;
}

#-------------------------------------------------------------------------
# void = ispec_str( \%instance_spec, $opt )
#
# Will return a string representing the instance spec.
#
# $opt: 'l', each component are printed on one line
#       '', selected components are printed all on one line 
#
sub ispec_str {

  my ($inst_spec, $opt) = @_;
  my $buf = '';

  if ( $opt && $opt =~ /l/ ) {
    foreach ( sort keys %$inst_spec ) {
      $buf .= "$_ : $inst_spec->{$_}\n";
    }
  }
  else {
    $buf = ispec2str( $inst_spec, 'fqgz' );
  }

  return $buf;
}

#-------------------------------------------------------------------------
# \%instance_spec = str2ispec( $instance_string )
#
# It will build a instance spec (hash) from a instance string.
# it's tailored to parse output from 'ups depend'. 
# meaning: it's very primitive, looking for only a few options, and expect 
# formats like:
# '<product> <version> -f <flavor> -q <qualifiers> -z database -g <chain>' or
# '-f <flavor> -q <qualifiers> <product> <version> -z database -g <chain>'.
#
sub str2ispec {

  my $inst_str = shift;
  my %inst_spec = ();
  
  @_ = split ' ', $inst_str;
  while( ($_ = shift) ) {
    s/\A"?(.*?)"?\Z/$1/o;
    if ( /^-f/ ) { $inst_spec{flavor} = shift; }
    elsif ( /^-q/ ) { $inst_spec{qualifiers} = shift; }
    elsif ( /^-z/ ) { $inst_spec{database} = shift; }
    elsif ( /^-g/ ) { $inst_spec{chain} = shift; }
    elsif ( /^-/ ) { shift; next; }
    elsif ( ! exists $inst_spec{product} ) { $inst_spec{product} = $_; }
    elsif ( ! exists $inst_spec{version} ) { $inst_spec{version} = $_; }
  }

  # a unique instance needs qualifiers set 

  $inst_spec{qualifiers} = "" unless ( exists $inst_spec{qualifiers} );

  return \%inst_spec;
}

#-------------------------------------------------------------------------
# $ret_code = ups_oxc( $ups_cmd, \@stdout_arr, \@stderr_arr )
#
# It will excute local ups commands
#
# Note: it return 0 on success, the ups return code on a ups error, 
#       else -1.
#       it will always buffer separate stderr and stdout, to mimic 
#       remote ups.
#
# Input : scalar, ups_command (without 'ups')
# Output: array ref, data, if not passed, data will go to stdout
#         array ref, errors, if not passed, errors will go to stderr
# Return: OS return code from ups
#
sub ups_oxc {

  my ($ups_cmd, $stdout_arr, $stderr_arr) = @_;
  my $ret_code = 0;
  my $err_file = "$$.ups_err";

  unless ( $ups_cmd ) {
    upderr_log( 0, "ERROR: Empty ups command\n" );
    return -1;
  }

  # always redirect stderr to a file

   if ( defined $ENV{TMPDIR} ) { 
     $err_file = "$ENV{TMPDIR}/$err_file"; 
   }
   else { 
     $err_file = "/tmp/$err_file"; 
   }
   unlink $err_file if ( -e $err_file );

   unless ( open( ERRFILE, ">$err_file" ) ) {
     upderr_log( 0, "ERROR: could not open temp file '$err_file': $!\n" );
     return -1;
   }

  close ERRFILE;

  # execute ups

  $ups_cmd = "ups $ups_cmd";  

  if ( defined $stdout_arr ) {
    @$stdout_arr = split /\n/, `$ups_cmd 2>> $err_file`;
  }
  else {
    print `$ups_cmd 2>> $err_file`;
  }
  $ret_code = $?;
  
  # read stderr
    
  if ( open( ERRFILE, $err_file ) ) {
    while ( <ERRFILE> ) { 
      chomp;
      if ( defined $stderr_arr ) { push @$stderr_arr, $_; }
      else { print $_; }
    }
    close ERRFILE;
  }
  else {
    upderr_log( 0, "ERROR: could not open temp file '$err_file': $!\n" );
    $ret_code = -1 unless ( $ret_code );
  }

  unlink $err_file;

  return $ret_code;
}

##########################################################################
# BEGIN/END
#
# Like to put'm here so everything else are defined.
#

sub BEGIN {
}

##########################################################################
# If standalone, run tests
#

if ( $0 =~ /updups.pm$/ ) {
  my $test_usage = "usage: perl updups.pm -h <upd_distribution_url> [-vvv]\n";

  my %a_inst = (
		product     => 'exmh',
		flavor      => 'IRIX',
		qualifiers  => '',
		prod_dir    => '/a/prod/dir',
		chain       => 'current',
		);

  my %x_arg = (
	       chain        => 'current',
	       prod_dir     => '/a/changed/prod/dir',
	       );

  my $rhost = '';
  my $verbose = 0;
  $g_flags{opt_s} = '1';

  # parse arguments

  while( ($_ = shift) ) {
    if ( /^-h/ ) { $rhost = shift; }
    elsif ( /^-v/ ) { $verbose = length( $_ ) -1; }
    else { print $test_usage; exit 1; }
  }

  # we need this one

  if ( ! $rhost ) {
    print $test_usage;
    exit 1;
  }

  upderr_setverbose( $verbose );

  print "\nTesting updups.pm\n";
  print "Local database        : $ENV{PRODUCTS}\n";
  print "upd distribution node : $rhost\n";

  # test local and remote ups calls 

  foreach my $url ('local',"$rhost") {

    my %b_inst = %a_inst; 
    my @arr = ();
    my $val = '';

    print "\nDoing ups on \'$url\':\n";
    
    if ( updups_flavor( $url, '', \$val ) ) {
      print "os flavor on $url: $val\n";
    }
    else {
      print "ERROR from updups_flavor\n";
    }

    $b_inst{version} = "v99_99";
    print "updups_exist:: ", ispec_str( \%b_inst ), "\n";
    if ( updups_exist( $url, \%b_inst, \$val ) ) {
      if ( $val ) {
	print "it does exist: ", ispec_str( \%b_inst ), "\n";
      }
      else {
	print "it does NOT exist: ", ispec_str( \%b_inst ), "\n";
      }
    }
    else {
      print "ERROR from updups_exist\n";
      next;
    }

    delete $b_inst{version};
    print "updups_find: ", ispec_str( \%b_inst ), "\n";
    if ( updups_find( $url, \%b_inst, 'version' ) ) {
      if ( defined $b_inst{version} ) {
	print "found: ", ispec_str( \%b_inst ), "\n";
      }
      else {
	print "NOT found: ", ispec_str( \%b_inst ), "\n";
      }
    }
    else {
      print "ERROR from updups_find\n";
      next;
    }

    print "updups_depend: ", ispec_str( \%b_inst ), "\n";
    if ( updups_depend( $url, \%b_inst, \@arr ) ) {
      if ( @arr ) {
	print "found:\n";
	foreach( @arr ) {
	  print "   ", ispec_str( $_ ), "\n";
	}
      }
      else {
	print "Product has not dependcies\n";
      }
    }
    else {
      print "ERROR from updups_depend\n";
    }

    print "updups_declare: ", ispec_str( \%b_inst ), "\n";
    updups_declare( $url, \%b_inst, \%x_arg );

  }
  
  print "\nError/Information stack:\n";
  upderr_get(); upderr_clear();
}

1;

__END__

=head1 NAME

 updups - This module contains routines to encapsulate ups on the local and remote host.


=head1 SYNOPSIS

=head1 DESCRIPTION

 This module contains routines to encapsulate ups on the local and remote host.

=head1 BUGS

 it's not all done yet

=cut
