##########################################################################
#
# FILE:
#       upd
# 
# DESCRIPTION: 
#       UPD main routine
#       (for more info, see pod at the end).
#
# AUTHORS:
#       Marc Mengel
#       Lars Rasmussen
#	Paul Russo
#       Margaret Votava
#       Don Walsh
#
#       Fermilab Computing Division
#       Batavia, IL 60510, U.S.A.                                              
#
# MODIFICATIONS:
#	29-jul-1998, llc, modify upduti_print_instance_spec to print
#	                  all key/value pairs in the instance hash
#	02-Feb-1998, psr, first
#	10-Feb-1998, psr, second
#
##########################################################################
package upduti;

use upderr;
use updgbl;
use strict;
use vars qw( @ISA @EXPORT );
use Exporter ();
use Cwd;

@ISA = qw( Exporter );
@EXPORT = qw(
	     upduti_print_instance_spec
	     upduti_parse_url
	     upduti_split_archive_file
	     upduti_get_proprietarylist
	     upduti_make_tarfile
	     upduti_expandtildes
	     );


##########################################################################
# Exported functions
#

#-------------------------------------------------------------------------
# status = upduti_print_instance_spec( $verbosity, \%instance_spec )
#
# prints out an instance spec for debugging purposes
#
# Input : scalar, verbosity level
#         hash, instance_spec
# Output: none
# Return: scalar, status
#

sub upduti_print_instance_spec {
    my ($verbosity, $spec ) = @_;
    my $keyword;

    my @FIXED_KEYWORDS = ( 
                "version",
                "flavor",
                "qualifiers",
                "chain",
                "database"
	 );

    my $LISTED_KEYWORDS = "|product|" . join("|", @FIXED_KEYWORDS) . "|";



#
#   print the first few keywords in a known, predictable order with
#   the "usual" stuff on top.
#
    upderr_log ($verbosity,"product name = \"$spec->{product}\"\n") if (defined ($spec->{product}));
    foreach $keyword ( @FIXED_KEYWORDS ) {
	upderr_log ($verbosity,"        $keyword = \"$spec->{\"$keyword\"}\"\n") if (defined ($spec->{"$keyword"}));
    }

#
#   print everything else alphabetically (if it hasn't already been printed)
#
    foreach $keyword ( sort(keys(%{$spec})) ) {
      if ( ! ("$LISTED_KEYWORDS" =~ m/"|$keyword|"/) ) {
	upderr_log ($verbosity,"        $keyword = \"$spec->{\"$keyword\"}\"\n") if (defined ($spec->{"$keyword"}));
      }
    }
return 1;
}

#-------------------------------------------------------------------------
# \%url_hash = upduti_parse_url( $url_str )
#
# It will parse an URL string and create a hash with the components:
# (protocol, node, port, path, str).
#
# Note: default protocol is http
#       default port is 80 if protocol is http, else empty
#       default path is empty
#
# Input : scalar, url string
# Output: none
# Return: hash ref, success
#         undef, on failer
#
sub upduti_parse_url {

  my $URL = shift;
  my @part = ();

  # default url

  my %url = (
	     protocol => '',
	     node     => '',
	     port     => '',
	     path     => '',
	     str      => ''
	     );
  
  # attempt to parse

  unless ( (@part = $URL =~ m@((\w+)://|)([^/:]+)(:\d*)?([^\#]*)@o) ) {
    upderr_log( 0, "ERROR: Failed to parse url: '$URL'\n" );
    return undef;
  }

  shift @part;

  # at least we need a node

  unless ( $part[1] ) {
    upderr_log( 0, "ERROR: No node defined in: '$URL'\n" );
    return undef;
  }

  # default protocol is http
    
  $part[0] = "http" unless ( defined $part[0] );

  # remove colon from from port number

  $part[2] =~ s/^:// if ( defined $part[2] );

  # for http, default port is 80

  $part[2] = 80 if ( $part[0] =~ /http/i && ! $part[2] );

  # make a hash

  my $n = 0;
  foreach ('protocol', 'node', 'port', 'path') {
    $url{$_} = $part[$n++];
  }
  
  # rebuild the string

  $url{str} = $part[0] . '://' . $part[1] . ':' . $part[2] . $part[3];

  return \%url;
}

sub upduti_split_archive_file {
    my ($url) = @_;
    my @list;

    # split into:
    #                    proto  :// host   dir  file  . suffix [.gz|Z]
    @list = ($url =~ m{\A([^:]*)://([^/:]*)(:\d+)?(/.*)/(.+?)\.(\w+(\.Z|\.gz)?)(\.\d+)?\Z}o );

    upderr_log(3, "split $url into " . join("|", @list) . "\n");
    return @list;
}

sub upduti_make_tarfile {
    my ($dir,$prodname) = @_;
    local $upduti::count;
    my ($files,$tarfile);
    my $cwd = cwd();

    $prodname="" if (!defined $prodname);
    upderr_log(3,"Entering dir is $dir...");
    if (!defined($upduti::count)) {
	$upduti::count = 0;
    }
    if (!defined $ENV{TMPDIR} || ! -f $ENV{TMPDIR}) {
	$ENV{TMPDIR} = "/tmp";
    }
    $tarfile = "$ENV{TMPDIR}/upd${$}_${upduti::count}.tar";
    $upduti::count++;

    unless ( eval {chdir($dir);} ) {
	upderr_log(0,"While trying to make tarfile:\n");
	upderr_log(0,"Could not chdir to $dir\n");
	return '';
    }

  if ($g_flags{opt_s}) {
     upderr_log(0,"I would make a tarfile of ${dir} called $tarfile now");
     eval {chdir($cwd);};
     return $tarfile;
  } else {

    $files = `sh -c "find . \\( \\( -name CVS -o -name SCCS \\) -prune \\) -o \\! -type d ! -name '.manifest*' -print 2>/dev/null | sort"`;

    if ( $files eq '' ) {
	upderr_log(0,"While trying to make tarfile:\n");
	upderr_log(0,"Could not generate file list of product home\n");
        eval {chdir($cwd);};
	return '';
    }
  
    upderr_log(4, "file list:\n$files");

    # make .manifest file

    open (OUT,"|xargs -n25 sum -r 2>/dev/null | sed -e 's/[ 	].*[ 	 ]/	/' > .manifest.${prodname}");
    print OUT $files;
    close(OUT);

    if ( ! -f ".manifest.${prodname}" ) {
	upderr_log(0,"While trying to make tarfile:\n");
	upderr_log(0,"Could not make .manifest file, continuing...");
    }

    # make initial tarfile with manifest
    system ("tar cf $tarfile .manifest.${prodname}");

    open(OUT, "|xargs -n25 tar uf $tarfile");
    print OUT $files;
    close(OUT);

    upderr_log(3,"Returning...");
    eval {chdir($cwd);};
    return $tarfile;
  }
}

sub upduti_expandtildes {
    my($string) = $_[0];
    my($users, @users, $rest, $user);
    my(@res) = ();
    my $dirsep = '/';
    my $debugflag = 0;

    # check for and do  tilde expansion
    #
    if ( $string =~ /\A\~([^${dirsep}]*)(.*)/ ) {
        $users = $1;
        $rest = $2;

        if ( $users =~ /\{.*\}/) {
            @users = ($users =~ m/(\w+)/g);
            print "here users is @users\n" if ($debugflag);
            foreach $user (@users) {
                unshift(@res, (&gethome($user) . $rest) );
            }

        } else {
            $string =~ s/\A\~([^${dirsep}]*)/&gethome($1)/e;
            unshift(@res, $string);
        }
    } else {
 	unshift(@res,$string);
    }
    return @res;
}

sub gethome {
    my(@list);
    if ( $_[0] eq '' ) {
        return $ENV{'HOME'};
    } else {
        @list = getpwnam($_[0]);
        return $list[7];
    }
}

##########################################################################
# Private functions
#

# If standalone, run tests
if ( $0 =~ /upduti.pm$/ )
{
my $err;
my %a_inst = (
           product     => "exmh",
           flavor      => "IRIX",
           version     => 'v1_3',
           qualifiers  => '',
           database    => "blah",
           foundby     => "current"
           );
upduti_print_instance_spec(0,\%a_inst);
upderr_get(); upderr_clear();

print "\n\n";
my $url = upduti_parse_url( 'http://fndapg.fnal.gov:8000/upd/ups' );
foreach ( sort keys %$url ) {
  print "$_ = $url->{$_}\n";
}
print "\n";
$url = upduti_parse_url( 'fndapg' );
foreach ( sort keys %$url ) {
  print "$_ = $url->{$_}\n";
}

}

1;

__END__

=head1 NAME

upd - blah blah

=head1 SYNOPSIS

<use it this way>

=head1 DESCRIPTION

<description>

=head1 BUGS

<bugs>

=cut
