##########################################################################
#
# FILE:
#       updxfr.pm
# 
# DESCRIPTION: 
#       File transfer routines for upd
#
# AUTHORS:
#       Marc Mengel
#       Lars Rasmussen
#       Margaret Votava
#       Don Walsh
#
#       Fermilab Computing Division
#       Batavia, Il 60510, U.S.A.                                              
#
# MODIFICATIONS:
#       dd-mmm-yyyy, NN, first
#
##########################################################################
package updxfr;

use strict;
use Exporter();


use vars qw( @ISA @EXPORT);
@ISA = qw(Exporter);

# export this !
@EXPORT = qw (
      updxfr_ls
      updxfr_file
      updxfr_dir
      updxfr_put_unique
);

##########################################################################
# Imported Modules
use English;
use FileHandle;
use File::Copy qw(copy);
use File::Path;
use Net::Domain qw(hostfqdn);
use Net::FTP;
use upderr;
use updgbl;

##########################################################################
# Exported variables
#

##########################################################################
# Global and local variables
#
# use vars qw();

# my g_<global variables at file scope> = '' # initialization

##########################################################################
# Exported functions
#

#-------------------------------------------------------------------------
# <return> = updxfr_ls( node, dir , fd )
#
# <description>
#
# Input : nodename "node", directory path "dir"
# Output: filehandle to list upon "fd"
# Return: <return>
#
sub updxfr_ls
{
    my( $node,			# node name to do list upon
        $dir,			# directory to chdir to/ls
        $fd,			# file descriptor to put list upon
	$prop_info		# proprietary info, if any
    ) = @_;			

    my($ftp);			# Net::FTP ftp handle
    my(@list);			# list of files returned
    my($f);			# individual entry from above list
 
    $ftp = &login_and_cwd( $node, $dir, $prop_info );

    if (!defined($ftp)) {
	return 0;
    }

    @list = $ftp->dir("-R");

    if (!$ftp->ok()) {
	&upderr_log( 3 , "ftp DIR command failed\n");
        $ftp->close;
	return;
    }

    if (!defined($fd)) {
        foreach $f (@list) {
	    print $f,"\n";
	}
    } else {
        foreach $f (@list) {
	    print $fd $f , "\n";
	}
    }
    $ftp->close;
}

sub updxfr_file
{
    my( $node,			# distribution node name to transfer from
        $dir,			# directory to chdir to there
        $dist, 			# file name on distribution node
        $local, 		# file name on local filesystem
	$prop_info		# proprietary login info
    ) = @_;			
    my( $ftp,		 	# Net::FTP server handle
	$osize,			# file size of file before transfer
	$size,			# file size of file after transfer
	$retry,			# we should retry as long as we get data
	$res,			# return code for this routine
        $local_dir,		# directory of local file
	$msg			# error message from things

    );

    &upderr_log( 3 , "updxfr_file($node,$dir,$dist,$local\n");

    # make directory for file if needed.

    ($local_dir) = ($local =~ m|(.*)/|o );

    if (defined($local_dir) && $local_dir ne "") {

	if ( defined($g_flags{opt_s}) && $g_flags{opt_s} ) {
	    upderr_log (0,"I would make directory $local_dir\n");
	} else {    
	  if (! -d $local_dir ) {
	    if (! eval {mkpath($local_dir, 0, 0777);} ) {
		upderr_log(0,"Unable to make directory $local_dir: $OS_ERROR\n");
		return 0;
	    }
	  }
	}
    }

    # initially we think the file is empty

    $osize = $size = 0;

    # I can no longer think why we wouldn't want to retry a transfer...
    # but we'll leave a variable here in case I do -- mengel
    $retry = 1;

    if ( defined($g_flags{opt_s}) && $g_flags{opt_s} ) {
        upderr_log (0,"I would fetch $dist as $local from $node now\n");
	$res = 1;
    } else {
        $ftp = &login_and_cwd($node,$dir, $prop_info );
	if ($ftp) {
	    
	    if ( -f $local) {
	 	$size = (stat($local))[7];
	    } else {
                $size = 0;
            }
	    $res = 1;
	    do {
                my $localfh = new FileHandle;
		&upderr_log( 3 , "setting binary transfer mode\n");
		$osize = $size;

		$ftp->binary;
		&upderr_log( 3 , "starting transfer at offset $size\n");

		$ftp->get( $dist, $local, $size );
		if ($ftp->ok()) {
		    $res = 1;
		} else {
		    $res = 0;
                    $msg = $ftp->message();
	            &upderr_log( 0 , "error: while attempting to ftp to $node:\n");
		    &upderr_log( 0, "\tftp transfer failed, reason $msg\n");
		}
		&upderr_log( 3 , "finished transfer result $res\n");
		$size = (stat($local))[7];
	    } until ( !$retry || $res == 1 || $size <= $osize );
            $ftp->quit;
	} else {
	   $res = 0;
	   $msg = $OS_ERROR;
	   &upderr_log( 0 , "error:while attempting to ftp to $node:\n");
	   &upderr_log( 0 , "\tftp connection failed: $OS_ERROR, \n");
	}
	unless (-f $local) {
	  $res = 0;
	  $msg = "unknown failure -- file did't exist after attempt";
	  upderr_log (0, "$local did not transfer\n"); 
	}
	if ($res) {
	    upderr_syslog( 'notice' , 
		"successful transfer ftp://${node}${dir}/${dist} -> $local\n" );
	    upderr_log(0, "informational: transferred ${dist}\n\tfrom ${node}:${dir} to\n\t${local}");
	} else {
	    upderr_syslog( 'warning' , 
			"failed transfer ftp://${node}${dir}/${dist} -> $local\n" );
	    upderr_syslog( 'warning' , "reason: $msg\n");
	    upderr_log(0, "error: can't transfer ${dir}/${dist}\n\tfrom ${node} to\n\t${local}");
	}
    }
    upderr_log( 3 , "returns $res\n");
    return $res;
}

sub updxfr_put_unique
{
    my( $node,			# distribution node name to transfer from
        $dir,			# directory to chdir to there
        $dist, 			# ref to return file name on distribution node
        $local, 		# file name on local filesystem
	$prop_info		# proprietary login info
    ) = @_;			
    my( $ftp,		 	# Net::FTP server handle
	$res,			# result code
	$msg,			# error message string
    );

    &upderr_log( 3 , "updxfr_put($node,$dir,$dist,$local\n");
    if ( ! -f $local ) {
        upderr_log(0,"cannot put file $local, file does not exist!\n");
	return 0;
    }

    if ( defined($g_flags{opt_s}) && $g_flags{opt_s} ) {
        upderr_log (0,"I would upload $local on $node now\n");
	$res = 1;
    } else {
        $ftp = &login_and_cwd($node,$dir, $prop_info );
	if ($ftp) {
	    
	    &upderr_log( 3 , "setting binary transfer mode\n");
	    $ftp->binary;

	    $$dist = $ftp->put_unique( $local );
	    if ($ftp->ok()) {
		$res = 1;
	    } else {
		$msg = $ftp->message();
		$res = 0;
		&upderr_log( 0, "error: ftp transfer failed, reason $msg\n");
	    }
	    &upderr_log( 3 , "finished transfer result $res name $$dist\n");
            $ftp->quit;
	} else {
	   $msg = $OS_ERROR;
	   &upderr_log( 0 , "ftp connection failed: $msg, \n");
	   $res = 0;
	}
    }
    &upderr_log( 3 , "returns $res\n");
    return $res;
}

%updxfr::can_pipe_into = ( 
     'tar' => 1, 
     'tar.gz' => 1, 
     'zip' => 0, 
     'exe' => 0, 
     'shar' => 0 
);

%updxfr::unpack_command = ( 
    'tar'    => 'tar xBf %s', 
    'tar.gz' => 'gunzip -c %s | tar xBf -', 
    'zip'    => 'unzip %s',
    'exe'    => '%s',
    'shar'   => 'sh %s'
);

%updxfr::verbose_unpack_command = ( 
    'tar'    => 'tar xvBf %s', 
    'tar.gz' => 'gunzip -c %s | tar xvBf -', 
    'zip'    => 'unzip %s',
    'exe'    => '%s',
    'shar'   => 'sh -x %s'
);

sub updxfr_dir
{
    my( $node,
        $dir,
        $dist, 
        $local, 
	$prop_info,		# proprietary login info
	$arch_type		# archive type (tar, zip, etc.)
    ) = @_;			# arguments, as above
    my( $ftp,			# Net::FTP server handle
	$out,			# output filehandle
	$res,			# return code for this routine
	$intermediate,		# use an intermediate file
	$unpack_cmd,		# command to unwind archive
	$msg			# error message string
    );

    $res = 0;

    &upderr_log( 3 , "updxfr_dir($node,$dir,$dist,$local\n");
    &upderr_log( 3 , "            $prop_info, $arch_type)\n");

    if ( defined($g_flags{opt_s}) && $g_flags{opt_s} ) {
        upderr_log (0,"I would make directory $local\n");
    } else {    
      if ( ! -d $local ) {
	if ( !eval {mkpath($local,0,0777);} ) {
             &upderr_log(0 , "unable to make directory $local: $OS_ERROR\n");
	     return 0;
	}
      }
    }

    if ( !defined($arch_type) || $arch_type eq "" ) {
	$arch_type = "tar";
    }
    upderr_log (3,"arch_type is $arch_type\n");

    # use an intermediate file if we were told to, or if
    # the archive type can't be piped into

    $intermediate = ($g_flags{opt_I} || !($updxfr::can_pipe_into{$arch_type}));

    if ( defined($g_flags{opt_s}) && $g_flags{opt_s} ) {
        upderr_log (0,"I would fetch directory $dist from ftp://$node/$dir as $local now\n"); 
	$res = 1;
    } else {
        $ftp = &login_and_cwd( $node, $dir, $prop_info );
	if ($ftp) {

	    &upderr_log( 3 , "setting binary transfer mode\n");
	    $ftp->binary;
	    &upderr_log( 3 , "starting transfer\n");
	    if (! chdir($local)) {
	       upderr_log(0,"Could not chdir to $local\n");
	       return 0;
            }
            if ($intermediate) {
		   
		    &upderr_log(3,"Using intermediate file\n");
		    # use an intermediate file to support retries...

		    # fetch it to destination.tar
		    $res = updxfr_file($node,	$dir,
					$dist . "." . $arch_type, 
			  		$local . "." . $arch_type);

		    # blow it into our tar to unwind it.
		    # we could catch unwind errors on the copy() or the close()

		    if (upderr_getverbose() > 0) {
			$unpack_cmd = 
			    sprintf($updxfr::verbose_unpack_command{$arch_type},
				    "$local.$arch_type" );
                    } else {
			$unpack_cmd = 
			    sprintf($updxfr::unpack_command{$arch_type},
				    "$local.$arch_type" );
		    }

		    if (0 != system("$unpack_cmd")) {
			$res = 0;
			$msg = $!;
			&upderr_log(0,"unwind failed: $msg\n");
		    }

		    # intermediate file is there for next time.  
		    unlink("$local.$arch_type");
		    
		} else {

		    # ftp transfer it directly into the tar to unwind.
		    #      
	            $out = new FileHandle; 

		    upderr_log(3, "current verbosity appears to be " , upderr_getverbose, "\n");

		    if (upderr_getverbose() > 0) {
			$unpack_cmd = 
			    sprintf($updxfr::verbose_unpack_command{$arch_type}, "-");
 		    } else {
			$unpack_cmd = 
			    sprintf($updxfr::unpack_command{$arch_type}, "-");
		    }

		    &upderr_log(3,"Piping data into $unpack_cmd\n");

		    my $handler = $::SIG{'PIPE'};
		    $::SIG{'PIPE'} = 'IGNORE';
		    		
	            if ( !$out->open("|$unpack_cmd")) {
		        $res = 0;
			$msg = $OS_ERROR
		        &upderr_log(0 , "starting unwind pipeline failed, $msg\n");
		    }
		    binmode $out;
     
		    $ftp->get( "${dist}.$arch_type", $out );
		    if ($ftp->ok()) {
			$res = 1;
		 	&upderr_log( 3 , "completed transfer\n");
		    } else {
			$res = 0;
			$msg = $ftp->message();
			&upderr_log( 0 , "error: ftp transfer failed: $msg \n");
		    }
		    unless (close($out)) {

			# this fails sometimes when it ought not to, so 
			# for now it doesn't return the error

			# $res = 0; 
			$msg = $!;
			if ($msg ne '') {
			   &upderr_log(0,"unwind appeared to fail?: $msg\n");
			}
		    }
		    $::SIG{'PIPE'} = $handler;
		}
            $ftp->quit;
	}
	if ($res) {
	    upderr_syslog( 'notice' , 
		"successful unwind ftp://${node}${dir}/${dist} -> $local\n");
		upderr_log(0, "informational: transferred ${dir}/${dist}\n\tfrom ${node} to\n\t${local}");
	} else {
	    upderr_syslog( 'warning' , 
		"failed unwind ftp://${node}${dir}/${dist} -> $local\n");
	    upderr_syslog( 'warning' , "reason: $msg\n");
	    upderr_log(0, "error: can't transfer ${dir}/${dist}\n\t from ${node} to \n\t${local}");
	}
    }
    &upderr_log( 3 , "returns $res\n");
    return $res;
}

##########################################################################
# Private functions
#

#-------------------------------------------------------------------------
# <return> = login_and_cwd( node name $node, directory path $dir , 
#		propinfo %propinfo)
#
# ftp's to the host, cd's to the directory, and returns the server handle
#
# Input : $node and $dir
# Output: none
# Return: Net::FTP handle
#
sub login_and_cwd {

    my( $node, 
        $dir,
	$prop_info		# proprietary login info
      ) = @_;

    my($ftp);
    my($actual);
    my($msg);

    upderr_log(1,"args node $node dir $dir\n");
    $ftp = Net::FTP->new($node);

    if (defined($ftp)) {
        upderr_log(3,"Attempting to login and cwd\n");
	$ftp->login('anonymous', $ENV{USER} . '@' . &hostfqdn());
	if (!$ftp->ok()) {
	    $msg = $ftp->message();
	    &upderr_log( 0 , "while attempting to ftp to $node:\n");
	    &upderr_log( 0 , "ftp login failed: $msg \n");
	    $ftp->close();
	    return undef;
	}

	if ( defined($prop_info->{group_name}) ) {
            upderr_log(3,"sending SITE GROUP \n");
	    $ftp->quot( "SITE" , "GROUP", $prop_info->{group_name});
	    if ( ! ($ftp->ok()) ) {
		$msg = $ftp->message();
	        &upderr_log( 0 , "while attempting to ftp to $node:\n");
		&upderr_log( 0 , "ftp SITE GROUP failed: $msg \n");
	        $ftp->close();
	        return undef;
	    }
	}

	if ( defined($prop_info->{pass_word} )) {
            upderr_log(3,"sending SITE GPASS \n");
	    $ftp->quot( "SITE" , "GPASS", $prop_info->{pass_word});
	    if (!$ftp->ok()) {
		$msg = $ftp->message();
		&upderr_log( 0 , "ftp SITE GPASS failed: $msg \n");
	        $ftp->close();
	        return undef;
	    }
	}

	$ftp->cwd($dir);

	if (!$ftp->ok()) {
	    $msg = $ftp->message();
	    &upderr_log( 0 , "while attempting to ftp to $node:\n");
	    &upderr_log( 0 , "ftp CWD failed: $msg \n");
	    $ftp->close();
	    return undef;
	}
    } else {
        upderr_log(0,"Unable to create ftp connection\n");
    }
    return $ftp;
}

# If standalone, run tests
if ( $0 =~ /updxfr.pm$/ ) {

   my @prop_info;
   @prop_info = (
	{}, 
	{'group_name' => 'upd' , 'pass_word' => 'foo' }
    );

   upderr_setverbose(9);

   my $errstr;

   if ( $#ARGV == -1 ) {

       print "default regression test:\n";
       print "listing /KITS on ftp.fnal.gov\n";

       &updxfr_ls('ftp.fnal.gov','/KITS', $main::STDOUT, $prop_info[0]);
       upderr_get(\$errstr); print STDERR $errstr;

       &updxfr_ls('ftp.fnal.gov','/KITS', $main::STDOUT, $prop_info[1]);
       upderr_get(\$errstr); print STDERR $errstr;

       print "fetching .proplogfile from ftp.fnal.gov\n";
       &updxfr_file('ftp.fnal.gov','/KITS','.proplogfile','testfile',$prop_info[0]);
       upderr_get(\$errstr); print STDERR $errstr;

       print "fetching bogus file from ftp.fnal.gov\n";
       &updxfr_file('ftp.fnal.gov','/KITS','nosuchfile','testfile2',$prop_info[0]);
       upderr_get(\$errstr); print STDERR $errstr;

       print "fetching .proplogfile to bogus place from  ftp.fnal.gov\n";
       &updxfr_file('ftp.fnal.gov','/KITS','.proplogfile','/nosuchdir/cantwritefile',$prop_info[0]);
       upderr_get(\$errstr); print STDERR $errstr;

       print "fetching and unwinding tarfile from ftp.fnal.gov\n";
       &updxfr_dir('ftp.fnal.gov','/KITS/GENERIC_UNIX/ups_compile/v1_3','ups_compilev1_3', 'testdir',$prop_info[0]);
       upderr_get(\$errstr); print STDERR $errstr;

       print "Done.\n"
    } elsif ( $ARGV[0] eq "-f" ) {
       shift(@ARGV);
       print "fetching $ARGV[0]/$ARGV[1]  as $ARGV[1]\n";
       &updxfr_file('ftp.fnal.gov', $ARGV[0], $ARGV[1], $ARGV[2],$prop_info[0]);
       print "Done.\n";
    } elsif ( $ARGV[0] eq "-i" ) {
        my $host = "ftp" ;
	my $dir = "/";
	my $dist = "README";
	my $local = "/tmp/junk";

       print "> "; 
       while (<STDIN>) {
	   if (/local *= *(.*)/) { $local = $1; }
	   elsif (  /dir *= *(.*)/) {   $dir = $1; }
	   elsif (  /cd *(.*)/)     {   $dir .= "/" . $1; }
	   elsif ( /dist *= *(.*)/) {  $dist = $1; }
	   elsif ( /host *= *(.*)/) {  $host = $1; }
	   elsif ( /^xf$|file/ ) { updxfr_file($host,$dir,$dist,$local,$prop_info[0]);}
	   elsif ( /^xd$|dir/ ) { updxfr_dir($host,$dir,$dist,$local,$prop_info[0]);}
	   elsif ( /^xp$|put_unique/ ) { updxfr_put_unique($host,$dir,\$dist,$local,$prop_info[0]);}
	   elsif ( /^xl$|ls/ ) { updxfr_ls($host,$dir);}
	   elsif ( /^quit|exit/) { exit 0; }
	   else { print "unkown command: $_\n";}
	   upderr_get(); upderr_clear();

	   print "local: $local\n";
	   print "host: $host\n";
	   print "dir: $dir\n";
	   print "dist: $dist\n";
           print "> "; 
       }
    }
}
1;

__END__

=head1 NAME

updxfr - blah blah

=head1 SYNOPSIS

<use it this way>

=head1 DESCRIPTION

<description>

=head1 BUGS

<bugs>

=cut
